/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_camera.h"
#include "lw_timer.h"
using namespace lwlib;

#define lwlib_Camera_KeyAction(name) lwlib_Camera_KeyAction_ ## name

#define lwlib_Camera_KeyAction_MoveLeft "a"
#define lwlib_Camera_KeyAction_MoveRight "d"
#define lwlib_Camera_KeyAction_MoveForward "w"
#define lwlib_Camera_KeyAction_MoveBackward "s"
#define lwlib_Camera_KeyAction_MoveUp "space"
#define lwlib_Camera_KeyAction_MoveDown "c"

#define lwlib_Camera_KeyAction_TurnLeft "h"
#define lwlib_Camera_KeyAction_TurnRight "l"
#define lwlib_Camera_KeyAction_LookUp "k"
#define lwlib_Camera_KeyAction_LookDown "j"

Camera_KeyToDir_t s_keyToDir[] =
{
	{ 
		lwlib_Camera_KeyAction(MoveLeft), 
		lwlib_vec3_init(0, 1, 0)
	},
	{ 
		lwlib_Camera_KeyAction(MoveRight),
		lwlib_vec3_init(0, -1, 0)
	},
	{
		lwlib_Camera_KeyAction(MoveForward),
		lwlib_vec3_init(1, 0, 0)
	},
	{
		lwlib_Camera_KeyAction(MoveBackward),
		lwlib_vec3_init(-1, 0, 0)
	},
	{
		lwlib_Camera_KeyAction(MoveUp),
		lwlib_vec3_init(0, 0, 1)
	},
	{
		lwlib_Camera_KeyAction(MoveDown),
		lwlib_vec3_init(0, 0, -1)
	},
};

void Camera::processMoveKey(Camera_KeyToDir_t *keyToDir)
{
	Mat3f rot;
	Vector3f move;

	if (lwlib_KeyDown(keyToDir->btnStr))
	{
		rot = rotationMat();
		if (priv.projectionType == Camera_ProjectionType::Orthographic)
		{
			rot = mat_mult(mat3f_rot(90, 0), rot);
		}

		move = vec_scale(keyToDir->dir, priv.moveSpeed * 
			g_frameTimer.secondsElapsed());
		priv.pos = vec_add(priv.pos, mat_xform(rot, move));
	}
}

Camera::Camera(lwlib_Keys_t keys, Mouse_t mouse)
{
	lwlib_Zero(priv);

	priv.keys = keys;
	priv.mouse = mouse;
	priv.moveSpeed = 16.0;
	priv.turnSpeedKeyboard = 90.0; // deg/sec
	priv.turnSpeedMouse = 1.0 / 100.0; // deg/sec/mouse unit
}

void Camera::setProjectionType(Camera_ProjectionType::e projectionType)
{
	priv.projectionType = projectionType;
}

void Camera::setAllowInput(bool allowInput)
{
	priv.allowInput = allowInput;
}

void Camera::setPosition(Point3f pos)
{
	priv.pos = pos;
}

Point3f Camera::getPosition()
{
	return priv.pos;
}

void Camera::setAngles(Point3f angles)
{
	priv.angles = angles;
}

Point3f Camera::getAngles()
{
	return priv.angles;
}

void Camera::updateFrame()
{
	double yawInc, yawIncSign;
	double pitchInc, pitchIncSign;
	Point2i mouseMove;

	if (priv.allowInput)
	{
		if (priv.projectionType == Camera_ProjectionType::Perspective)
		{
			mouseMove = priv.mouse->getMove();

			yawIncSign = 0;
			if (lwlib_KeyDown(lwlib_Camera_KeyAction_TurnLeft))
			{
				yawIncSign = 1;
			}
			else if (lwlib_KeyDown(lwlib_Camera_KeyAction_TurnRight))
			{
				yawIncSign = -1;
			}
			yawInc = yawIncSign * priv.turnSpeedKeyboard * 
				g_frameTimer.secondsElapsed();

			if (lwlib_X(mouseMove) != 0)
			{
				yawInc -= lwlib_X(mouseMove) * priv.turnSpeedMouse;
			}

			pitchIncSign = 0;
			if (lwlib_KeyDown(lwlib_Camera_KeyAction_LookUp))
			{
				pitchIncSign = 1;
			}
			else if (lwlib_KeyDown(lwlib_Camera_KeyAction_LookDown))
			{
				pitchIncSign = -1;
			}
			pitchInc = pitchIncSign * priv.turnSpeedKeyboard * 
				g_frameTimer.secondsElapsed();

			if (lwlib_Y(mouseMove) != 0)
			{
				pitchInc -= lwlib_Y(mouseMove) * priv.turnSpeedMouse;
			}

			priv.angles = vec_add(priv.angles, vec3f(0, pitchInc, yawInc));
			lwlib_Y(priv.angles) = lwlib_CLIP(lwlib_Y(priv.angles), -90.0, 90.0);
		}

		lwlib_RepeatArray(s_keyToDir, i, keyToDir,
			processMoveKey(&keyToDir)
			);
	}
}

Vector3f Camera::getDirection()
{
	return mat_xform(rotationMat(), vec3f(1.0, 0.0, 0.0));
}

Mat3f Camera::rotationMat()
{
	Mat3f pitchRot, yawRot;

	pitchRot = mat3f_rot(-lwlib_Y(priv.angles), 1);
	yawRot = mat3f_rot(lwlib_Z(priv.angles), 2);

	return mat_mult(yawRot, pitchRot);
}

void Camera::addToEdit(Edit *edit)
{
	edit->addItem_double("moveSpeed", &priv.moveSpeed);
	edit->addItem_double("turnSpeedKeyboard", &priv.turnSpeedKeyboard);
	edit->addItem_double("turnSpeedMouse", &priv.turnSpeedMouse);
	lwlib_Edit_CheckPersist(priv.persist, addToEdit(fsEdit));
}
