/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_edit.h"
#include "lw_strmap.h"
#include "lw_font.h"
#include "lw_timer.h"
using namespace lwlib;

#define lwlib_Edit_DoubleChangeSlow 0.01
#define lwlib_Edit_DoubleChangeFast 1.0

#define lwlib_Edit_IntChangeSlow 1
#define lwlib_Edit_IntChangeFast 10

#define lwlib_Edit_KeyAction(name) lwlib_Edit_KeyAction_ ## name
#define lwlib_Edit_KeyAction_ToggleEditBox ";"
#define lwlib_Edit_KeyAction_EnterGroup "e"
#define lwlib_Edit_KeyAction_ExitGroup "u"
#define lwlib_Edit_KeyAction_NextItem "j"
#define lwlib_Edit_KeyAction_PrevItem "k"
#define lwlib_Edit_KeyAction_ToggleBool "t"
#define lwlib_Edit_KeyAction_Increase "a"
#define lwlib_Edit_KeyAction_Decrease "z"
#define lwlib_Edit_KeyAction_IncreaseFast "shift-a"
#define lwlib_Edit_KeyAction_DecreaseFast "shift-z"
#define lwlib_Edit_KeyAction_ToggleSign "-"
#define lwlib_Edit_KeyAction_CursorDelete "x"
#define lwlib_Edit_KeyAction_CursorLeft "h"
#define lwlib_Edit_KeyAction_CursorRight "l"
#define lwlib_Edit_KeyAction_CursorBackspace "bksp"
#define lwlib_Edit_KeyAction_ClearValue "c"
#define lwlib_Edit_KeyAction_EditModeString "e"
#define lwlib_Edit_KeyAction_EditModeNormal "enter"
#define lwlib_Edit_KeyAction_CopyValue "y"
#define lwlib_Edit_KeyAction_PasteValue "p"
#define lwlib_Edit_KeyAction_AppendText "shift-a"
#define lwlib_Edit_KeyPressed(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_Edit_KeyAction(name), lwlib_Keys_Event(Pressed))
#define lwlib_Edit_KeyReleased(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_Edit_KeyAction(name), lwlib_Keys_Event(Released))
#define lwlib_Edit_KeyRepeat(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_Edit_KeyAction(name), lwlib_Keys_Event(Repeat))
#define lwlib_Edit_KeyFastRepeat(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_Edit_KeyAction(name), lwlib_Keys_Event(FastRepeat))
#define lwlib_Edit_KeyDown(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_Edit_KeyAction(name), lwlib_Keys_Event(Down))

#define lwlib_Edit_AccessItem(item, type) (*(item->type ## Value))

#define lwlib_Edit_CursorChar "\x0b"
#define lwlib_Edit_CursorFlashDuration 0.150

#define lwlib_Edit_IndentSize 2

#define LWLIB_EDIT_CHECKGROUPVEC(item) \
	assert( \
		(item) != NULL && \
		(item)->type == Edit_ItemType::Group && \
		(item)->groupType == Edit_ItemGroupType::Vec && \
		(item)->groupVec != NULL && \
		(item)->parentItem != NULL && \
		(item)->parentItem->type == Edit_ItemType::Group && \
		(item)->parentItem->groupType == Edit_ItemGroupType::Normal \
		)

#define LWLIB_EDIT_DEPTHSPACING 4

typedef lwlib_Byte_t byte;

static int lwlib_Edit_Item_Index(Edit_Item_t *item, const char *itemName)
{
	int i;

	for (i = 0; (size_t)i < vec_size(item->groupItems); i++)
	{
		if (strcmp(vec_at(item->groupItems, i)->name, itemName) == 0)
		{
			return i;
		}
	}

	return -1;
}

Edit_Item_t *Edit::findItem(const char *itemName)
{
	int itemIndex;
	Edit_Item_t *item;

	assert(priv.nav->type == Edit_ItemType::Group);

	item = NULL;
	itemIndex = lwlib_Edit_Item_Index(priv.nav, itemName);

	if (itemIndex != -1)
	{
		item = vec_at(priv.nav->groupItems, itemIndex);
	}

	return item;
}

static bool lwlib_Edit_Item_Exists(Edit_Item_t *item, const char *itemName)
{
	return lwlib_Edit_Item_Index(item, itemName) != -1;
}

static void lwlib_Edit_Item_Clear(Edit_Item_t *item)
{
	int i;
	Edit_Item_t *subItem;

	for (i = 0; (size_t)i < vec_size(item->groupItems); i++)
	{
		subItem = vec_at(item->groupItems, i);
		lwlib_Edit_Item_Clear(subItem);
		free(subItem);
	}
	vec_free(item->groupItems, Edit_ItemHandle_t);
	free(item->groupVec);

	lwlib_Zero(*item);
}

Edit_Item_t *Edit::addItem(const char *itemName, 
	Edit_ItemType::e itemType)
{
	int itemIndex;
	Edit_Item_t *newItem;
	Edit_Item_t *item;

	newItem = NULL;
	assert(priv.nav->type == Edit_ItemType::Group);

	if (lwlib_Edit_Item_Exists(priv.nav, itemName))
	{
		itemIndex = lwlib_Edit_Item_Index(priv.nav, itemName);
		newItem = vec_at(priv.nav->groupItems, itemIndex);
		lwlib_Edit_Item_Clear(newItem);
		lwlib_StrCopy(newItem->name, itemName);
		newItem->type = itemType;
		newItem->parentItem = priv.nav;
		newItem->itemGroupIndex = itemIndex;
	}
	else
	{
		item = lwlib_CallocSingle(Edit_Item_t);
		lwlib_StrCopy(item->name, itemName);
		item->type = itemType;
		item->parentItem = priv.nav;
		item->itemGroupIndex = priv.nav->groupItemCount++;
		vec_push(priv.nav->groupItems, Edit_ItemHandle_t, item);

		newItem = vec_top(priv.nav->groupItems);
	}

	assert(newItem != NULL);
	return newItem;
}

Edit_Item_t *Edit::addItem_WithValue(const char *itemName, 
	Edit_ItemType::e itemType, void *valueData,
	int valueLength)
{
	Edit_Item_t *newItem;
	newItem = addItem(itemName, itemType);
	newItem->valueData = (byte *)valueData;
	newItem->valueLength = valueLength;
	return newItem;
}

void Edit::resetItemSelectors()
{
	int i;
	lwlib_StrSeq_t path;
	lwlib_Str_t itemName;

	navigateTop();
	lwlib_Zero(path);

	for (i = 0; (size_t)i < vec_size(priv.groupItemPath); i++)
	{
		itemName = vec_at(priv.groupItemPath, i);
		if (navigateItem(itemName.v))
		{
			vec_push(path, lwlib_Str_t, itemName);
		}
		else
		{
			break;
		}
	}

	vec_free(priv.groupItemPath, lwlib_Str_t);
	lwlib_Copy(priv.groupItemPath, path);

	if (!lwlib_Edit_Item_Exists(priv.nav, priv.selectedItemName))
	{
		lwlib_StrCopy(priv.selectedItemName, "");
		if (vec_size(priv.nav->groupItems) > 0)
		{
			lwlib_StrCopy(priv.selectedItemName, 
				vec_at(priv.nav->groupItems, 0)->name);
		}
	}

	navigateTop();
}

bool Edit::navigateItemByPath(lwlib_StrSeq_t itemPath)
{
	int i;

	for (i = 0; (size_t)i < vec_size(itemPath); i++)
	{
		if (!navigateItem(vec_at(itemPath, i).v))
		{
			return false;
		}
	}

	return true;
}

void Edit::item_DisplayString(Edit_Item_t *item, 
	lwlib_String_t dispStr)
{
	int index;
	int enumValue;
	Plane3f tPlane3fValue;
	Sphere3f tSphere3fValue;
	lwlib_String_t groupValueStr;

	switch (item->type)
	{
	case Edit_ItemType::Int:
		lwlib_VaString(dispStr, "%s %d", 
			item->name, lwlib_Edit_AccessItem(item, int));
		break;
	case Edit_ItemType::Short:
		lwlib_VaString(dispStr, "%s %d", 
			item->name, (int)lwlib_Edit_AccessItem(item, short));
		break;
	case Edit_ItemType::UChar:
		lwlib_VaString(dispStr, "%s %d", 
			item->name, (int)lwlib_Edit_AccessItem(item, uchar));
		break;
	case Edit_ItemType::Double:
		lwlib_VaString(dispStr, "%s %.2f", 
			item->name, lwlib_Edit_AccessItem(item, double));
		break;
	case Edit_ItemType::String:
		lwlib_VaString(dispStr, "%s \"%s\"", 
			item->name, item->stringValue);
		break;
	case Edit_ItemType::Bool:
		lwlib_VaString(dispStr, "%s %s", 
			item->name, lwlib_Edit_AccessItem(item, bool) ? "true" : "false");
		break;
	case Edit_ItemType::Enum:
		enumValue = lwlib_Edit_AccessItem(item, enum);
		if (item->enumValues != NULL)
		{
			index = lwlib_IndexOf(enumValue, item->enumValues, 
				item->enumNumValues);
			if (index != -1)
			{
				lwlib_VaString(dispStr, "%s 0x%X:%s", 
					item->name, item->enumValues[index], 
					item->enumValueNames[index]);
			}
			else
			{
				lwlib_VaString(dispStr, "%s", item->name);
			}
		}
		else
		{
			assert(enumValue >= 0 && enumValue < item->enumNumValues);
			lwlib_VaString(dispStr, "%s %d:%s", item->name, 
				enumValue, item->enumValueNames[enumValue]);
		}
		break;
	case Edit_ItemType::Group:
		lwlib_StrCopy(groupValueStr, "");
		if (item->groupType == Edit_ItemGroupType::TPoint3f)
		{
			lwlib_VaString(groupValueStr, "(%.2f, %.2f, %.2f)", 
				lwlib_ExpandVector(lwlib_Edit_AccessItem(item, tPoint3f)));
			lwlib_VaString(dispStr, "%s %s", item->name, groupValueStr);
		}
		else if (item->groupType == Edit_ItemGroupType::TPoint2f)
		{
			lwlib_VaString(groupValueStr, "(%.2f, %.2f)", 
				lwlib_ExpandVector(lwlib_Edit_AccessItem(item, tPoint2f)));
			lwlib_VaString(dispStr, "%s %s", item->name, groupValueStr);
		}
		else if (item->groupType == Edit_ItemGroupType::TPoint2i)
		{
			lwlib_VaString(groupValueStr, "(%d, %d)", 
				lwlib_ExpandVector(lwlib_Edit_AccessItem(item, tPoint2i)));
			lwlib_VaString(dispStr, "%s %s", item->name, groupValueStr);
		}
		else if (item->groupType == Edit_ItemGroupType::TPlane3f)
		{
			tPlane3fValue = lwlib_Edit_AccessItem(item, tPlane3f);
			lwlib_VaString(groupValueStr, "(%.2f, %.2f, %.2f) (%.2f, %.2f, %.2f)", 
				lwlib_ExpandVector(lwlib_plane_origin(tPlane3fValue)),
				lwlib_ExpandVector(lwlib_plane_normal(tPlane3fValue))
				);
			lwlib_VaString(dispStr, "%s %s", item->name, groupValueStr);
		}
		else if (item->groupType == Edit_ItemGroupType::TSphere3f)
		{
			tSphere3fValue = lwlib_Edit_AccessItem(item, tSphere3f);
			lwlib_VaString(groupValueStr, "(%.2f, %.2f, %.2f) %.2f", 
				lwlib_ExpandVector(tSphere3fValue.origin),
				tSphere3fValue.radius
				);
			lwlib_VaString(dispStr, "%s %s", item->name, groupValueStr);
		}
		else if (item->groupType == Edit_ItemGroupType::Vec)
		{
			LWLIB_EDIT_CHECKGROUPVEC(item);
			lwlib_VaString(dispStr, "%s[%d:%d]", item->name, 
				item->groupVec->selector, vec_size(*(item->groupVec->core)));
		}
		else if (item->groupType == Edit_ItemGroupType::VecElem)
		{
			lwlib_VaString(dispStr, "%s", item->name);
		}
		else
		{
			lwlib_VaString(dispStr, "%s", item->name);
		}

		break;
	}
}

bool Edit::item_Selected(Edit_Item_t *item)
{
	return strcmp(item->name, priv.selectedItemName) == 0 ||
		(priv.selectedItemName[0] == '\0' && item->itemGroupIndex == 0);
}

int Edit::item_CursorCol(Edit_Item_t *item)
{
	int cursorCol;
	int intValue;
	double doubleValue;
	lwlib_String_t tmpStr;
	Point2i colRange;

	if (!item_Selected(item))
	{
		return -1;
	}

	cursorCol = -1;
	switch (item->type)
	{
	case Edit_ItemType::Int:
		intValue = lwlib_Edit_AccessItem(item, int);
		lwlib_Snprintf(tmpStr, "%d", intValue);
		colRange = vec2i(intValue < 0 ? 1 : 0, strlen(tmpStr));
		cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);
		break;
	case Edit_ItemType::Short:
		intValue = (int)lwlib_Edit_AccessItem(item, short);
		lwlib_Snprintf(tmpStr, "%d", intValue);
		colRange = vec2i(intValue < 0 ? 1 : 0, strlen(tmpStr));
		cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);
		break;
	case Edit_ItemType::UChar:
		intValue = (int)lwlib_Edit_AccessItem(item, uchar);
		lwlib_Snprintf(tmpStr, "%d", intValue);
		colRange = vec2i(0, strlen(tmpStr));
		cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);
		break;
	case Edit_ItemType::Double:
		doubleValue = lwlib_Edit_AccessItem(item, double);
		lwlib_Snprintf(tmpStr, "%.2f", doubleValue);
		colRange = vec2i(doubleValue < 0.0 ? 1 : 0, strlen(tmpStr));
		cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);
		break;
	case Edit_ItemType::String:
		cursorCol = lwlib_CLIP(priv.cursorCol, 0, strlen(item->stringValue));
		break;
	case Edit_ItemType::Bool:
		cursorCol = 0;
		break;
	case Edit_ItemType::Enum:
		cursorCol = 0;
		break;
	case Edit_ItemType::Group:
		break;
	}

	return cursorCol;
}

void Edit::drawCursor(Edit_Item_t *item, Point2i pos)
{
	int cursorCol, drawCol;
	double timeSince;

	cursorCol = item_CursorCol(item);
	if (cursorCol != -1)
	{
		if (priv.cursorDrawTimer.isZero())
		{
			priv.cursorDrawTimer = priv.timeNow;
		}
		timeSince = priv.cursorDrawTimer.gapInSeconds(priv.timeNow);
		if (timeSince >= 0.0 && timeSince < lwlib_Edit_CursorFlashDuration)
		{
			drawCol = strlen(item->name) + 1 + cursorCol;
			if (item->type == Edit_ItemType::String)
			{
				drawCol++;
			}

			pos = priv.font->advancePosFit(pos, drawCol);
			priv.font->printPosFit(pos, lwlib_Edit_CursorChar);
		}
		else
		{
			if (timeSince >= lwlib_Edit_CursorFlashDuration)
			{
				priv.cursorDrawTimer = priv.timeNow;
				priv.cursorDrawTimer.addDuration(
					lwlib_Edit_CursorFlashDuration);
			}
		}
	}
}

int Edit::queryIncDecKeys()
{
	int change;

	change = 0;
	if (lwlib_Edit_KeyRepeat(IncreaseFast))
	{
		change = lwlib_Edit_IntChangeFast;
	}
	else if (lwlib_Edit_KeyRepeat(DecreaseFast))
	{
		change = -lwlib_Edit_IntChangeFast;
	}
	else if (lwlib_Edit_KeyRepeat(Increase))
	{
		change = lwlib_Edit_IntChangeSlow;
	}
	else if (lwlib_Edit_KeyRepeat(Decrease))
	{
		change = -lwlib_Edit_IntChangeSlow;
	}

	return change;
}

bool Edit::item_HandleCopyPasteKeys(Edit_Item_t *item)
{
	bool ret = false;
	int regCode;
	Edit_CopyRegister_t *copyReg;

	if (lwlib_Edit_KeyDown(CopyValue) || lwlib_Edit_KeyPressed(CopyValue))
	{
		regCode = -1;
		if (lwlib_Edit_KeyPressed(CopyValue))
		{
			regCode = 0;
		}
		else
		{
			regCode = lwlib_Keys_FirstNumericEvent(priv.keys, lwlib_Keys_Event(Pressed));
		}

		if (regCode != -1)
		{
			copyReg = &priv.copyRegisters[regCode];
			copyReg->type = item->type;
			free(copyReg->valueData);
			copyReg->valueData = lwlib_CallocMany(byte, item->valueLength);
			memcpy(copyReg->valueData, item->valueData, item->valueLength);
			copyReg->valueLength = item->valueLength;
		}

		ret = true;
	}

	if (lwlib_Edit_KeyDown(PasteValue) || lwlib_Edit_KeyPressed(PasteValue))
	{
		regCode = -1;
		if (lwlib_Edit_KeyPressed(PasteValue))
		{
			regCode = 0;
		}
		else
		{
			regCode = lwlib_Keys_FirstNumericEvent(priv.keys, lwlib_Keys_Event(Pressed));
		}

		if (regCode != -1)
		{
			copyReg = &priv.copyRegisters[regCode];
			if (copyReg->type == item->type && copyReg->valueLength <= item->valueLength)
			{
				memset(item->valueData, 0, item->valueLength);
				memcpy(item->valueData, copyReg->valueData, copyReg->valueLength);
			}
		}

		ret = true;
	}

	return ret;
}

void Edit::item_HandleIntEditKeys(Edit_Item_t *item)
{
	int change;
	int intValue;
	int tmpValue;
	char *nextKey;
	bool tmpStrModified;
	Point2i colRange;
	lwlib_Str_t tmpStr;

	intValue = lwlib_Edit_AccessItem(item, int);

	change = queryIncDecKeys();
	if (change != 0)
	{
		intValue += change;
	}

	if (lwlib_Edit_KeyRepeat(ToggleSign))
	{
		intValue = -intValue;
		priv.cursorCol = 0;
	}
	if (lwlib_Edit_KeyRepeat(ClearValue))
	{
		intValue = 0;
		priv.cursorCol = 0;
	}

	lwlib_Snprintf(tmpStr.v, "%d", intValue);
	colRange = vec2i(intValue < 0 ? 1 : 0, strlen(tmpStr.v));
	priv.cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);

	if (lwlib_Edit_KeyRepeat(CursorLeft))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + vec_range(colRange) - 1) %
			vec_range(colRange);
	}
	else if (lwlib_Edit_KeyRepeat(CursorRight))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + 1) %
			vec_range(colRange);
	}

	tmpStrModified = false;
	if (lwlib_Edit_KeyRepeat(CursorDelete))
	{
		if (strlen(tmpStr.v) - lwlib_X(colRange) == 1)
		{
			tmpStr.v[priv.cursorCol] = '0';
		}
		else
		{
			tmpStr = lwlib_Str_DeleteChar(tmpStr, priv.cursorCol);
		}
		tmpStrModified = true;
	}

	lwlib_Keys_EventQueueReset(priv.keys, 
		lwlib_Keys_NumericBtnStrs(priv.keys), lwlib_Keys_Event(Repeat));

	while (lwlib_Keys_EventQueuePending(priv.keys) && !tmpStrModified)
	{
		nextKey = lwlib_Keys_EventQueueNext(priv.keys);

		if (strlen(tmpStr.v) - lwlib_X(colRange) == 1 &&
			tmpStr.v[priv.cursorCol] == '0')
		{
			tmpStr.v[priv.cursorCol] = *nextKey;
		}
		else
		{
			priv.cursorCol++;
			tmpStr = lwlib_Str_InsertStr(tmpStr, lwlib_Str(nextKey), priv.cursorCol);
		}
		tmpStrModified = true;
	}

	if (tmpStrModified)
	{
		tmpValue = 0;
		if (sscanf(tmpStr.v, "%d", &tmpValue) == 1)
		{
			intValue = tmpValue;
		}
	}
	
	lwlib_Edit_AccessItem(item, int) = intValue;
}

void Edit::item_HandleShortEditKeys(Edit_Item_t *item)
{
	short change;
	short shortValue;
	short tmpValue;
	char *nextKey;
	bool tmpStrModified;
	Point2i colRange;
	lwlib_Str_t tmpStr;

	shortValue = lwlib_Edit_AccessItem(item, short);

	change = queryIncDecKeys();
	if (change != 0)
	{
		shortValue += change;
	}

	if (lwlib_Edit_KeyRepeat(ToggleSign))
	{
		shortValue = -shortValue;
		priv.cursorCol = 0;
	}
	if (lwlib_Edit_KeyRepeat(ClearValue))
	{
		shortValue = 0;
		priv.cursorCol = 0;
	}

	lwlib_Snprintf(tmpStr.v, "%hi", shortValue);
	colRange = vec2i(shortValue < 0 ? 1 : 0, strlen(tmpStr.v));
	priv.cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);

	if (lwlib_Edit_KeyRepeat(CursorLeft))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + vec_range(colRange) - 1) %
			vec_range(colRange);
	}
	else if (lwlib_Edit_KeyRepeat(CursorRight))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + 1) %
			vec_range(colRange);
	}

	tmpStrModified = false;
	if (lwlib_Edit_KeyRepeat(CursorDelete))
	{
		if (strlen(tmpStr.v) - lwlib_X(colRange) == 1)
		{
			tmpStr.v[priv.cursorCol] = '0';
		}
		else
		{
			tmpStr = lwlib_Str_DeleteChar(tmpStr, priv.cursorCol);
		}
		tmpStrModified = true;
	}

	lwlib_Keys_EventQueueReset(priv.keys, 
		lwlib_Keys_NumericBtnStrs(priv.keys), lwlib_Keys_Event(Repeat));

	while (lwlib_Keys_EventQueuePending(priv.keys) && !tmpStrModified)
	{
		nextKey = lwlib_Keys_EventQueueNext(priv.keys);

		if (strlen(tmpStr.v) - lwlib_X(colRange) == 1 &&
			tmpStr.v[priv.cursorCol] == '0')
		{
			tmpStr.v[priv.cursorCol] = *nextKey;
		}
		else
		{
			priv.cursorCol++;
			tmpStr = lwlib_Str_InsertStr(tmpStr, lwlib_Str(nextKey), priv.cursorCol);
		}
		tmpStrModified = true;
	}

	if (tmpStrModified)
	{
		tmpValue = 0;
		if (sscanf(tmpStr.v, "%hi", &tmpValue) == 1)
		{
			shortValue = tmpValue;
		}
	}
	
	lwlib_Edit_AccessItem(item, short) = shortValue;
}

void Edit::item_HandleUCharEditKeys(Edit_Item_t *item)
{
	unsigned char change;
	unsigned char ucharValue;
	unsigned char tmpValue;
	char *nextKey;
	bool tmpStrModified;
	Point2i colRange;
	lwlib_Str_t tmpStr;

	ucharValue = lwlib_Edit_AccessItem(item, uchar);

	change = queryIncDecKeys();
	if (change != 0)
	{
		ucharValue += change;
	}

	if (lwlib_Edit_KeyRepeat(ClearValue))
	{
		ucharValue = 0;
		priv.cursorCol = 0;
	}

	lwlib_Snprintf(tmpStr.v, "%hhi", ucharValue);
	colRange = vec2i(0, strlen(tmpStr.v));
	priv.cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);

	if (lwlib_Edit_KeyRepeat(CursorLeft))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + vec_range(colRange) - 1) %
			vec_range(colRange);
	}
	else if (lwlib_Edit_KeyRepeat(CursorRight))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + 1) %
			vec_range(colRange);
	}

	tmpStrModified = false;
	if (lwlib_Edit_KeyRepeat(CursorDelete))
	{
		if (strlen(tmpStr.v) - lwlib_X(colRange) == 1)
		{
			tmpStr.v[priv.cursorCol] = '0';
		}
		else
		{
			tmpStr = lwlib_Str_DeleteChar(tmpStr, priv.cursorCol);
		}
		tmpStrModified = true;
	}

	lwlib_Keys_EventQueueReset(priv.keys, 
		lwlib_Keys_NumericBtnStrs(priv.keys), lwlib_Keys_Event(Repeat));

	while (lwlib_Keys_EventQueuePending(priv.keys) && !tmpStrModified)
	{
		nextKey = lwlib_Keys_EventQueueNext(priv.keys);

		if (strlen(tmpStr.v) - lwlib_X(colRange) == 1 &&
			tmpStr.v[priv.cursorCol] == '0')
		{
			tmpStr.v[priv.cursorCol] = *nextKey;
		}
		else
		{
			priv.cursorCol++;
			tmpStr = lwlib_Str_InsertStr(tmpStr, lwlib_Str(nextKey), priv.cursorCol);
		}
		tmpStrModified = true;
	}

	if (tmpStrModified)
	{
		tmpValue = 0;
		if (sscanf(tmpStr.v, "%hhi", &tmpValue) == 1)
		{
			ucharValue = tmpValue;
		}
	}
	
	lwlib_Edit_AccessItem(item, uchar) = ucharValue;
}

void Edit::item_HandleDoubleEditKeys(Edit_Item_t *item)
{
	double doubleChange;
	double doubleValue;
	double tmpValue;
	char *nextKey;
	bool tmpStrModified;
	Point2i colRange;
	lwlib_Str_t tmpStr;

	doubleValue = lwlib_Edit_AccessItem(item, double);

	doubleChange = 0.0;
	if (lwlib_Edit_KeyRepeat(IncreaseFast))
	{
		doubleChange = lwlib_Edit_DoubleChangeFast;
	}
	else if (lwlib_Edit_KeyRepeat(DecreaseFast))
	{
		doubleChange = -lwlib_Edit_DoubleChangeFast;
	}
	else if (lwlib_Edit_KeyRepeat(Increase))
	{
		doubleChange = lwlib_Edit_DoubleChangeSlow;
	}
	else if (lwlib_Edit_KeyRepeat(Decrease))
	{
		doubleChange = -lwlib_Edit_DoubleChangeSlow;
	}
	if (doubleChange != 0.0)
	{
		doubleValue += doubleChange;
	}

	if (lwlib_Edit_KeyRepeat(ToggleSign))
	{
		doubleValue = -doubleValue;
		priv.cursorCol = 0;
	}
	if (lwlib_Edit_KeyRepeat(ClearValue))
	{
		doubleValue = 0.0;
		priv.cursorCol = 0;
	}

	lwlib_Snprintf(tmpStr.v, "%.2f", doubleValue);
	colRange = vec2i(doubleValue < 0.0 ? 1 : 0, strlen(tmpStr.v));
	priv.cursorCol = lwlib_CLIP(priv.cursorCol, lwlib_X(colRange), lwlib_Y(colRange) - 1);

	if (lwlib_Edit_KeyRepeat(CursorLeft))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + vec_range(colRange) - 1) %
			vec_range(colRange);
		if (tmpStr.v[priv.cursorCol] == '.')
		{
			priv.cursorCol--;
		}
	}
	else if (lwlib_Edit_KeyRepeat(CursorRight))
	{
		priv.cursorCol = lwlib_X(colRange) + 
			((priv.cursorCol - lwlib_X(colRange)) + 1) %
			vec_range(colRange);
		if (tmpStr.v[priv.cursorCol] == '.')
		{
			priv.cursorCol++;
		}
	}

	tmpStrModified = false;
	if (lwlib_Edit_KeyRepeat(CursorDelete))
	{
		if (strchr(&tmpStr.v[priv.cursorCol], '.') != NULL)
		{
			if (lwlib_Str_IndexOf(tmpStr, '.') - lwlib_X(colRange) == 1)
			{
				tmpStr.v[priv.cursorCol] = '0';
			}
			else
			{
				tmpStr = lwlib_Str_DeleteChar(tmpStr, priv.cursorCol);
				if (tmpStr.v[priv.cursorCol] == '.')
				{
					priv.cursorCol--;
				}
			}
		}
		else
		{
			tmpStr.v[priv.cursorCol] = '0';
		}
		tmpStrModified = true;
	}

	lwlib_Keys_EventQueueReset(priv.keys, 
		lwlib_Keys_NumericBtnStrs(priv.keys), lwlib_Keys_Event(Repeat));

	while (lwlib_Keys_EventQueuePending(priv.keys) && !tmpStrModified)
	{
		nextKey = lwlib_Keys_EventQueueNext(priv.keys);

		if (strchr(&tmpStr.v[priv.cursorCol], '.') != NULL)
		{
			if (lwlib_Str_IndexOf(tmpStr, '.') - lwlib_X(colRange) == 1 &&
				tmpStr.v[priv.cursorCol] == '0')
			{
				tmpStr.v[priv.cursorCol] = *nextKey;
			}
			else
			{
				priv.cursorCol++;
				tmpStr = lwlib_Str_InsertStr(tmpStr, lwlib_Str(nextKey), priv.cursorCol);
			}
		}
		else
		{
			tmpStr.v[priv.cursorCol] = *nextKey;
			priv.cursorCol = lwlib_MIN(priv.cursorCol + 1, lwlib_Y(colRange) - 1);
		}
		tmpStrModified = true;
	}

	if (tmpStrModified)
	{
		tmpValue = 0.0;
		if (sscanf(tmpStr.v, "%lf", &tmpValue) == 1)
		{
			doubleValue = tmpValue;
		}
	}
	
	lwlib_Edit_AccessItem(item, double) = doubleValue;
}

void Edit::item_HandleStringEditKeys(Edit_Item_t *item)
{
	char nextKey;
	bool tmpStrModified;
	int range;
	lwlib_Str_t tmpStr;

	tmpStr = lwlib_Str(item->stringValue);
	tmpStrModified = false;

	priv.cursorCol = lwlib_CLIP(priv.cursorCol, 0, strlen(tmpStr.v));

	if (lwlib_Edit_KeyRepeat(CursorBackspace))
	{
		if (strlen(tmpStr.v) > 0 && priv.cursorCol > 0)
		{
			priv.cursorCol = lwlib_CLIP(priv.cursorCol - 1, 0, strlen(tmpStr.v) - 1);
			tmpStr = lwlib_Str_DeleteChar(tmpStr, priv.cursorCol);
			tmpStrModified = true;
		}
	}

	if (priv.editMode == Edit_EditMode::String)
	{
		lwlib_Keys_EventQueueReset(priv.keys, 
			lwlib_Keys_AsciiBtnStrs(priv.keys), lwlib_Keys_Event(Repeat));

		while (lwlib_Keys_EventQueuePending(priv.keys))
		{
			nextKey = lwlib_Keys_EventQueueNextAscii(priv.keys);

			if (strlen(item->stringValue) + 1 < (size_t)item->stringSize)
			{
				tmpStr = lwlib_Str_InsertStr(tmpStr, 
					lwlib_Str_FromChar(nextKey), priv.cursorCol);
				tmpStrModified = true;
				priv.cursorCol++;
			}
		}

		if (lwlib_Edit_KeyPressed(EditModeNormal))
		{
			priv.editMode = Edit_EditMode::Normal;
		}
	}
	else
	{
		if (lwlib_Edit_KeyRepeat(CursorLeft))
		{
			range = strlen(item->stringValue) + 1;
			priv.cursorCol = (priv.cursorCol + range - 1) % range;
		}
		else if (lwlib_Edit_KeyRepeat(CursorRight))
		{
			range = strlen(item->stringValue) + 1;
			priv.cursorCol = (priv.cursorCol + 1) % range;
		}

		if (lwlib_Edit_KeyRepeat(CursorDelete))
		{
			if (strlen(tmpStr.v) > 0)
			{
				priv.cursorCol = lwlib_CLIP(priv.cursorCol, 0, strlen(tmpStr.v) - 1);
				tmpStr = lwlib_Str_DeleteChar(tmpStr, priv.cursorCol);
				tmpStrModified = true;
			}
		}

		if (lwlib_Edit_KeyPressed(EditModeString))
		{
			priv.editMode = Edit_EditMode::String;
		}

		if (lwlib_Edit_KeyPressed(AppendText))
		{
			priv.editMode = Edit_EditMode::String;
			priv.cursorCol = strlen(tmpStr.v);
		}
	}

	if (tmpStrModified)
	{
		strncpy(item->stringValue, tmpStr.v, item->stringSize - 1);
		item->stringValue[item->stringSize - 1] = '\0';
	}
}

void Edit::item_HandleGroupEditKeys(Edit_Item_t *item)
{
	int vecsize;
	int change;
	lwlib_String_t tmpStr;

	if (lwlib_Edit_KeyRepeat(EnterGroup))
	{
		vec_push(priv.groupItemPath, lwlib_Str_t, lwlib_Str(priv.selectedItemName));
		if (item->groupType == Edit_ItemGroupType::Vec)
		{
			vecsize = vec_size(*(item->groupVec->core));
			if (vecsize > 0)
			{
				LWLIB_EDIT_CHECKGROUPVEC(item);
				item->groupVec->selector = lwlib_CLIP(
					item->groupVec->selector, 0, vecsize - 1);
				lwlib_VaString(tmpStr, "%s:%d", item->parentItem->name, item->groupVec->selector);
				vec_push(priv.groupItemPath, lwlib_Str_t, lwlib_Str(tmpStr));
			}
		}
	}

	if (item->groupType == Edit_ItemGroupType::Vec)
	{
		LWLIB_EDIT_CHECKGROUPVEC(item);
		change = queryIncDecKeys();
		if (change != 0)
		{
			vecsize = vec_size(*(item->groupVec->core));
			if (vecsize > 0)
			{
				item->groupVec->selector = (item->groupVec->selector + 
					vecsize + change) % vecsize;
			}
			else
			{
				item->groupVec->selector = 0;
			}
		}
	}
}

void Edit::vecChangeNotify()
{
	int selector;
	int elemSize;
	Edit_Item_t *vecItem;
	Edit_Vec_t tmpGroupVec;
	lwlib_VoidVec_t *core;
	lwlib_VoidVec_t newCore;
	lwlib_String_t navItemName;
	bool doInsert, doDelete;

	vecItem = findItem("elem");
	LWLIB_EDIT_CHECKGROUPVEC(vecItem);
	
	tmpGroupVec = *(vecItem->groupVec);
	core = tmpGroupVec.core;
	elemSize = tmpGroupVec.elemSize;
	selector = tmpGroupVec.selector;
	doInsert = tmpGroupVec.doInsert;
	doDelete = tmpGroupVec.doDelete;

	if (!(doInsert ? !doDelete : doDelete))
	{
		return;
	}
	vecItem->groupVec->doInsert = vecItem->groupVec->doDelete = false;

	lwlib_Zero(newCore);
	newCore.sz = core->sz;
	if (doInsert)
	{
		selector = lwlib_CLIP(selector, 0, core->sz);
		newCore.sz++;
	}
	else // doDelete
	{
		assert(doDelete);

		if (core->sz == 0)
		{
			return;
		}

		assert(core->sz > 0);
		selector = lwlib_CLIP(selector, 0, core->sz - 1);
		newCore.sz--;
	}
	newCore.data = vecI_reserve(elemSize, newCore.sz, 
		&newCore.alloc, newCore.data);
	
	if (doDelete)
	{
		memcpy(newCore.data, core->data, 
			selector * elemSize);

		if (tmpGroupVec.elemCleanup != NULL)
		{
			tmpGroupVec.elemCleanup((byte *)core->data + 
				selector * elemSize);
		}

		memcpy((byte *)newCore.data + selector * elemSize,
			(byte *)core->data + (selector + 1) * elemSize,
			(core->sz - selector - 1) * elemSize);
	}
	else // doInsert
	{
		assert(doInsert);

		memcpy(newCore.data, core->data, 
			selector * elemSize);

		memset((byte *)newCore.data + 
			selector * elemSize, 0,
			elemSize);
		if (tmpGroupVec.elemInit != NULL)
		{
			tmpGroupVec.elemInit(
				(byte *)newCore.data + selector * elemSize);
		}

		memcpy((byte *)newCore.data + (selector + 1) * elemSize, 
			(byte *)core->data + selector * elemSize, 
			(core->sz - selector) * elemSize);
	}

	core->data = vecI_reserve(elemSize, 0, 
		&core->alloc, core->data);
	core->sz = 0;
	*core = newCore;
	lwlib_Zero(newCore);

	lwlib_StrCopy(navItemName, priv.nav->name);
	navigateBack();
	addItem_vec(navItemName, core, elemSize, 
		tmpGroupVec.elemInit, tmpGroupVec.elemCleanup, 
		tmpGroupVec.elemAddToEdit);
	navigateItem(navItemName);

	if (doDelete)
	{
		if (selector > 0 && (size_t)selector == core->sz)
		{
			selector--;
		}
	}

	vecItem = findItem("elem");
	LWLIB_EDIT_CHECKGROUPVEC(vecItem);
	vecItem->groupVec->selector = selector;
}

int Edit::navItemDepth()
{
	int depth;
	Edit_Item_t *item;

	depth = 0;
	item = priv.nav;

	while (item->parentItem != NULL)
	{
		if (strcmp(item->name, "elem") != 0)
		{
			depth++;
		}
		item = item->parentItem;
	}

	return depth;
}

const char *Edit::navItemDepthSpacing(lwlib_String_t depthStr)
{
	int depth;

	depth = navItemDepth();
	depth *= LWLIB_EDIT_DEPTHSPACING;
	depth = lwlib_MIN(depth, lwlib_StringSize - 1);
	memset(depthStr, ' ', depth);
	depthStr[depth] = '\0';

	return depthStr;
}

Point2i Edit::drawGroupItems(Edit_Item_t *item, Point2i pos)
{
	int i;
	int oldIndent;
	int selector, vecSize;
	Edit_Item_t *subItem;
	Font_Color::e fontColor;
	lwlib_String_t dispStr;
	Point2i oldpos;
	bool selected;

	for (i = 0; (size_t)i < vec_size(item->groupItems); i++)
	{
		subItem = vec_at(item->groupItems, i);
		if (lwlib_BitOn(subItem->flags, Edit_ItemFlag::Disabled))
		{
			continue;
		}

		item_DisplayString(subItem, dispStr);

		selected = false;
		fontColor = Font_Color::Grey;
		if (item_Selected(subItem))
		{
			fontColor = Font_Color::Green;
			if (priv.editMode == Edit_EditMode::String)
			{
				fontColor = Font_Color::LightBlue;
			}
			selected = true;
		}

		priv.font->setColor(fontColor);

		oldpos = pos;
		pos = priv.font->printPosFit(pos, dispStr);
		drawCursor(subItem, oldpos);

		priv.font->setColor(Font_Color::Grey);

		if (selected)
		{
			lwlib_X(priv.selectedRange) = lwlib_Y(oldpos);
			lwlib_Y(priv.selectedRange) = lwlib_Y(pos);
			priv.gotSelectedRange = true;
		}

		if (subItem->type == Edit_ItemType::Group && 
			lwlib_BitOn(subItem->flags, Edit_ItemFlag::GroupExpand))
		{
			oldIndent = priv.font->getIndent();
			priv.font->setIndent(oldIndent + lwlib_Edit_IndentSize);
			pos = priv.font->advancePosFit(pos, -1);

			if (subItem->groupType == Edit_ItemGroupType::Vec)
			{
				vecSize = vec_size(*(subItem->groupVec->core));
				if (vecSize > 0)
				{
					selector = lwlib_CLIP(subItem->groupVec->selector, 0, vecSize - 1);
					pos = drawGroupItems(
						vec_at(subItem->groupItems, selector), pos);
				}
			}
			else
			{
				pos = drawGroupItems(subItem, pos);
			}
			priv.font->setIndent(oldIndent);
			lwlib_X(pos) -= lwlib_Edit_IndentSize;
		}
		else
		{
			pos = priv.font->advancePosFit(pos, -1);
		}
	}

	return pos;
}

void Edit::cycleSelectedItem(int cycle)
{
	Edit_Item_t *selectedItem;

	selectedItem = vec_at(priv.nav->groupItems, 
		lwlib_Edit_Item_Index(priv.nav, priv.selectedItemName));
	assert(lwlib_BitOff(selectedItem->flags, Edit_ItemFlag::Disabled));
	
	do
	{
		selectedItem = vec_at(priv.nav->groupItems, (selectedItem->itemGroupIndex + 
			cycle + priv.nav->groupItemCount) % priv.nav->groupItemCount);
	}
	while (lwlib_BitOn(selectedItem->flags, Edit_ItemFlag::Disabled));

	lwlib_StrCopy(priv.selectedItemName, selectedItem->name);
}

void Edit::scrollToSelected()
{
	int maxPos;

	if (priv.gotSelectedRange)
	{
		priv.gotSelectedRange = false;
		maxPos = priv.font->maxScreenPosY();

		if (lwlib_Y(priv.selectedRange) >= maxPos)
		{
			priv.scroll++;
		}
		else if 
			(
				lwlib_X(priv.selectedRange) < 0 || 
				(
					lwlib_X(priv.selectedRange) < maxPos / 2 && 
					priv.scroll > 0 && 
					priv.scroll < maxPos / 2 && 
					lwlib_X(priv.titleRange) < 0
				)
			)
		{
			priv.scroll--;
		}
	}
	else
	{
		if (priv.scroll > 0)
		{
			priv.scroll--;
		}
		else if (priv.scroll < 0)
		{
			priv.scroll++;
		}
	}
}

Edit::Edit()
{
	lwlib_Zero(priv);

	priv.topItem = lwlib_CallocSingle(Edit_Item_t);
	lwlib_StrCopy(priv.topItem->name, "top");
	priv.topItem->type = Edit_ItemType::Group;
	priv.nav = priv.topItem;
}

void Edit::newFs(const char *fileName, Fs_FileMode::e fileMode)
{
	Fs *fs;

	fs = new Fs(fileName, fileMode);
	priv.fs = fs;
	priv.fsOwner = true;
}

Edit::~Edit()
{
	int i;
	clearItems();
	if (priv.fsOwner)
	{
		delete priv.fs;
	}
	lwlib_Edit_Item_Clear(priv.topItem);
	free(priv.topItem);
	vec_free(priv.groupItemPath, lwlib_Str_t);
	for (i = 0; (size_t)i < lwlib_CountArray(priv.copyRegisters); i++)
	{
		free(priv.copyRegisters[i].valueData);
	}
}

void Edit::navigateTop()
{
	priv.nav = priv.topItem;
}

bool Edit::navigateItem(const char *itemName)
{
	bool found = false;

	if (priv.nav->type == Edit_ItemType::Group && 
		lwlib_Edit_Item_Exists(priv.nav, itemName))
	{
		priv.nav = vec_at(priv.nav->groupItems, 
			lwlib_Edit_Item_Index(priv.nav, itemName));
		found = true;
	}

	return found;
}

bool Edit::navigateBack()
{
	bool found = false;
	Edit_Item_t *oldNav;
	lwlib_String_t tmpStr, depthStr;

	oldNav = priv.nav;

	if (priv.nav->parentItem != NULL)
	{
		priv.nav = priv.nav->parentItem;
		found = true;
	}

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitBlock(priv.fs, NULL, 0, Fs_VarType::Informative,
		lwlib_VaString(tmpStr, "%s[%s]", depthStr, oldNav->name));

	return found;
}

void Edit::addItem_int(const char *itemName, int *value)
{
	Edit_Item_t *newItem;
	lwlib_String_t tmpStr, depthStr;

	newItem = addItem_WithValue(itemName, Edit_ItemType::Int,
		value, sizeof(int));
	newItem->intValue = value;

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitVar(priv.fs, value, Fs_VarType::Int, 
		lwlib_VaString(tmpStr, "%s%s", depthStr, itemName));
}

void Edit::addItem_short(const char *itemName, short *value)
{
	Edit_Item_t *newItem;
	lwlib_String_t tmpStr, depthStr;

	newItem = addItem_WithValue(itemName, Edit_ItemType::Short, 
		value, sizeof(short));
	newItem->shortValue = value;

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitVar(priv.fs, value, Fs_VarType::Short, 
		lwlib_VaString(tmpStr, "%s%s", depthStr, itemName));
}

void Edit::addItem_uchar(const char *itemName, unsigned char *value)
{
	Edit_Item_t *newItem;
	lwlib_String_t tmpStr, depthStr;

	newItem = addItem_WithValue(itemName, Edit_ItemType::UChar, 
		value, sizeof(unsigned char));
	newItem->ucharValue = value;

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitVar(priv.fs, value, Fs_VarType::UChar, 
		lwlib_VaString(tmpStr, "%s%s", depthStr, itemName));
}

void Edit::addItem_double(const char *itemName, double *value)
{
	Edit_Item_t *newItem;
	lwlib_String_t tmpStr, depthStr;

	newItem = addItem_WithValue(itemName, Edit_ItemType::Double, 
		value, sizeof(double));
	newItem->doubleValue = value;

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitVar(priv.fs, value, Fs_VarType::Double,
		lwlib_VaString(tmpStr, "%s%s", depthStr, itemName));
}

void Edit::addItem_string(const char *itemName, char *value, 
	int valueSize)
{
	int len;
	lwlib_String_t tmpStr, depthStr;
	Edit_Item_t *newItem;

	newItem = addItem_WithValue(itemName, Edit_ItemType::String,
		value, valueSize);
	newItem->stringValue = value;
	newItem->stringSize = valueSize;

	if (priv.fs != NULL &&
		priv.fs->getFileMode() == Fs_FileMode::ReadText)
	{
		len = valueSize;
	}
	else
	{
		len = strlen(value);
	}

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitBlock(priv.fs, value, len, Fs_VarType::String,
		lwlib_VaString(tmpStr, "%s%s", depthStr, itemName));

	if (priv.fs != NULL &&
		priv.fs->getFileMode() == Fs_FileMode::ReadText)
	{
		value[len - 1] = '\0';
	}
}

void Edit::addItem_bool(const char *itemName, bool *value)
{
	Edit_Item_t *newItem;
	lwlib_String_t tmpStr, depthStr;

	newItem = addItem_WithValue(itemName, Edit_ItemType::Bool,
		value, sizeof(bool));
	newItem->boolValue = value;
	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitVar(priv.fs, value, Fs_VarType::Bool, 
		lwlib_VaString(tmpStr, "%s%s", depthStr, itemName));
}

void Edit::addItem_enum(const char *itemName, int *value, 
	const char *valueNames[], const int *values, int numValues)
{
	Edit_Item_t *newItem;
	lwlib_String_t dispStr, tmpStr, depthStr;

	newItem = addItem_WithValue(itemName, Edit_ItemType::Enum,
		value, sizeof(int));
	newItem->enumValue = value;
	newItem->enumValueNames = valueNames;
	newItem->enumValues = values;
	newItem->enumNumValues = numValues;

	navItemDepthSpacing(depthStr);
	item_DisplayString(newItem, dispStr);
	lwlib_Fs_CommitVar(priv.fs, value, Fs_VarType::Int, 
		lwlib_VaString(tmpStr, "%s%s", depthStr, dispStr));
}

void Edit::addItem_group(const char *itemName)
{
	lwlib_String_t tmpStr, depthStr;

	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitBlock(priv.fs, NULL, 0, Fs_VarType::Informative,
		lwlib_VaString(tmpStr, "%s[%s]", depthStr, itemName));

	addItem(itemName, Edit_ItemType::Group);
	navigateItem(itemName);
}

void Edit::addItem_vec(const char *itemName, lwlib_VoidVec_t *value, 
	int elemSize, VecInit_t elemInit, VecCleanup_t elemCleanup, 
	VecAddToEdit_t elemAddToEdit)
{
	int i;
	int sz;
	Edit_Item_t *newItem;
	lwlib_String_t tmpStr, descStr, depthStr;

	assert(value != NULL && elemSize != 0);

	addItem_group(itemName);
	{
		Fs *oldFs = priv.fs;
		lwlib_Zero(priv.fs);

		addItem_group("elem");

		priv.fs = oldFs;
	}

	newItem = priv.nav;
	newItem->flags = lwlib_Pow2(Edit_ItemFlag::GroupExpand);
	newItem->groupType = Edit_ItemGroupType::Vec;
	newItem->groupVec = lwlib_CallocSingle(Edit_Vec_t);
	newItem->groupVec->core = value;
	newItem->groupVec->elemSize = elemSize;
	newItem->groupVec->elemInit = elemInit;
	newItem->groupVec->elemCleanup = elemCleanup;
	newItem->groupVec->elemAddToEdit = elemAddToEdit;

	sz = value->sz;
	lwlib_Snprintf(descStr, "%s:vecsize", itemName);
	navItemDepthSpacing(depthStr);
	lwlib_Fs_CommitVar(priv.fs, &sz, Fs_VarType::Int, 
		lwlib_VaString(tmpStr, "%s%s", depthStr, descStr));

	// handle growing or shrinking vector
	if ((size_t)sz != value->sz)
	{
		// clean up elements before deleting
		if ((size_t)sz < value->sz)
		{
			if (elemCleanup != NULL)
			{
				for (i = sz; (size_t)i < value->sz; i++)
				{
					elemCleanup((byte *)value->data + (elemSize * i));
				}
			}
		}

		// resize vector
		value->data = vecI_reserve(elemSize, sz, 
			&value->alloc, value->data);

		// init new elements
		if ((size_t)sz > value->sz)
		{
			memset((byte *)value->data + (elemSize * value->sz), 
				0, elemSize * (sz - value->sz));
			if (elemInit != NULL)
			{
				for (i = value->sz; i < sz; i++)
				{
					elemInit((byte *)value->data + (elemSize * i));
				}
			}
		}

		value->sz = sz;
	}

	for (i = 0; (size_t)i < value->sz; i++)
	{
		lwlib_Snprintf(tmpStr, "%s:%d", itemName, i);
		addItem_group(tmpStr);
		priv.nav->groupType = Edit_ItemGroupType::VecElem;

		if (elemAddToEdit != NULL)
		{
			elemAddToEdit((byte *)value->data + (elemSize * i), this);
		}

		navigateBack();
	}

	{
		Fs *oldFs = priv.fs;
		lwlib_Zero(priv.fs);

		navigateBack(); // leaving "elem"

		addItem_int("selector", &newItem->groupVec->selector);
		addItem_bool("doInsert", &newItem->groupVec->doInsert);
		addItem_bool("doDelete", &newItem->groupVec->doDelete);

		setItemFlag("doInsert", Edit_ItemFlag::VecChange, true);
		setItemFlag("doDelete", Edit_ItemFlag::VecChange, true);

		priv.fs = oldFs;
	}

	navigateBack(); // leaving "<itemName>"
}

void Edit::addItem_TPoint3f(const char *itemName, Point3f *value)
{
	addItem_group(itemName);
	priv.nav->groupType = Edit_ItemGroupType::TPoint3f;
	priv.nav->tPoint3fValue = value;
	addItem_double("x", &value->v[0]);
	addItem_double("y", &value->v[1]);
	addItem_double("z", &value->v[2]);
	navigateBack();
}

void Edit::addItem_TPoint2f(const char *itemName, Point2f *value)
{
	addItem_group(itemName);
	priv.nav->groupType = Edit_ItemGroupType::TPoint2f;
	priv.nav->tPoint2fValue = value;
	addItem_double("x", &value->v[0]);
	addItem_double("y", &value->v[1]);
	navigateBack();
}

void Edit::addItem_TPoint2i(const char *itemName, Point2i *value)
{
	addItem_group(itemName);
	priv.nav->groupType = Edit_ItemGroupType::TPoint2i;
	priv.nav->tPoint2iValue = value;
	addItem_int("x", &value->v[0]);
	addItem_int("y", &value->v[1]);
	navigateBack();
}

void Edit::addItem_TPlane3f(const char *itemName, Plane3f *value)
{
	addItem_group(itemName);
	priv.nav->groupType = Edit_ItemGroupType::TPlane3f;
	priv.nav->tPlane3fValue = value;
	addItem_TPoint3f("origin", &value->p[0]);
	addItem_TPoint3f("normal", &value->p[1]);
	navigateBack();
}

void Edit::addItem_TBox3f(const char *itemName, Box3f *value)
{
	addItem_group(itemName);
	priv.nav->groupType = Edit_ItemGroupType::TBox3f;
	priv.nav->tBox3fValue = value;
	addItem_TPoint3f("lower", &value->p[0]);
	addItem_TPoint3f("upper", &value->p[1]);
	navigateBack();
}

void Edit::addItem_TSphere3f(const char *itemName, Sphere3f *value)
{
	addItem_group(itemName);
	priv.nav->groupType = Edit_ItemGroupType::TSphere3f;
	priv.nav->tSphere3fValue = value;
	addItem_TPoint3f("origin", &value->origin);
	addItem_double("radius", &value->radius);
	navigateBack();
}

void Edit::clearItems()
{
	lwlib_Edit_Item_Clear(priv.topItem);
	lwlib_StrCopy(priv.topItem->name, "top");
	priv.topItem->type = Edit_ItemType::Group;
}

int Edit::getItemFlags(const char *itemName)
{
	int itemFlags;
	Edit_Item_t *item;

	itemFlags = 0;

	item = findItem(itemName);
	if (item != NULL)
	{
		itemFlags = item->flags;
	}

	return itemFlags;
}

void Edit::setItemFlags(const char *itemName, int itemFlags)
{
	Edit_Item_t *item;

	item = findItem(itemName);
	if (item != NULL)
	{
		item->flags = itemFlags;
	}
}

void Edit::setItemFlag(const char *itemName, Edit_ItemFlag::e itemFlag, 
	bool enabled)
{
	int flg;
	flg = getItemFlags(itemName);
	lwlib_SetBitLValue(flg, itemFlag, enabled);
	setItemFlags(itemName, flg);
}

void Edit::draw()
{
	int i;
	Point2i pos;
	const char *pathStr;
	Edit_Item_t *item;

	if (priv.active == false)
	{
		return;
	}

	pos = vec2i(0, 0);
	lwlib_Y(pos) -= priv.scroll;

	lwlib_X(priv.titleRange) = lwlib_Y(pos);
	pos = priv.font->printPosFit(pos, "top");

	navigateTop();
	for (i = 0; (size_t)i < vec_size(priv.groupItemPath); i++)
	{
		pathStr = "->";
		pos = priv.font->printPosFit(pos, pathStr);

		pathStr = vec_at(priv.groupItemPath, i).v;
		navigateItem(pathStr);
		pos = priv.font->printPosFit(pos, pathStr);
	}
	item = priv.nav;
	navigateTop();

	lwlib_Y(priv.titleRange) = lwlib_Y(pos);

	pos = vec2i(0, lwlib_Y(pos) + 1);
	pos = drawGroupItems(item, pos);

	scrollToSelected();
}

void Edit::updateFrame()
{
	int i, c;
	int index;
	bool status;
	Edit_Item_t *selectedItem;
	int enumChange;
	int enumValue;
	bool forceExitStringMode;
	lwlib_StrSeq_t itemPath;

	priv.timeNow = Timer();

	if (priv.editMode == Edit_EditMode::Normal)
	{
		if (lwlib_Edit_KeyPressed(ToggleEditBox))
		{
			priv.active = !priv.active;
		}
	}

	if (priv.active == false)
	{
		return;
	}

	resetItemSelectors();

	navigateTop();
	lwlib_Copy(itemPath, priv.groupItemPath);
	status = navigateItemByPath(itemPath);
	assert(status == true);

	assert(priv.selectedItemName[0] == '\0' || 
		lwlib_Edit_Item_Exists(priv.nav, priv.selectedItemName));

	selectedItem = NULL;
	if (priv.selectedItemName[0] != '\0')
	{
		selectedItem = vec_at(priv.nav->groupItems, 
			lwlib_Edit_Item_Index(priv.nav, priv.selectedItemName));
	}

	forceExitStringMode = false;
	if (priv.editMode == Edit_EditMode::String)
	{
		forceExitStringMode = true;
	}
	else
	{
		if (selectedItem != NULL)
		{
			if (item_HandleCopyPasteKeys(selectedItem))
			{
				selectedItem = NULL;
			}
		}
	}

	if (selectedItem != NULL)
	{
		switch (selectedItem->type)
		{
		case Edit_ItemType::Int:
			item_HandleIntEditKeys(selectedItem);
			break;
		case Edit_ItemType::Short:
			item_HandleShortEditKeys(selectedItem);
			break;
		case Edit_ItemType::UChar:
			item_HandleUCharEditKeys(selectedItem);
			break;
		case Edit_ItemType::Double:
			item_HandleDoubleEditKeys(selectedItem);
			break;
		case Edit_ItemType::String:
			item_HandleStringEditKeys(selectedItem);
			forceExitStringMode = false;
			break;
		case Edit_ItemType::Bool:
			if (lwlib_Edit_KeyRepeat(ToggleBool) || 
				lwlib_Edit_KeyRepeat(Increase) || lwlib_Edit_KeyRepeat(Decrease))
			{
				lwlib_Edit_AccessItem(selectedItem, bool) = 
					!lwlib_Edit_AccessItem(selectedItem, bool);
				if (lwlib_BitOn(selectedItem->flags, 
					Edit_ItemFlag::VecChange))
				{
					vecChangeNotify();
				}
			}
			break;
		case Edit_ItemType::Enum:
			enumValue = lwlib_Edit_AccessItem(selectedItem, enum);

			enumChange = 0;
			if (lwlib_Edit_KeyRepeat(Increase))
			{
				enumChange = 1;
			}
			else if (lwlib_Edit_KeyRepeat(Decrease))
			{
				enumChange = -1;
			}
			if (enumChange != 0)
			{
				if (selectedItem->enumValues != NULL)
				{
					index = lwlib_IndexOf(enumValue, selectedItem->enumValues,
						selectedItem->enumNumValues);
					if (index != -1)
					{
						index = (index + selectedItem->enumNumValues + enumChange) % 
							selectedItem->enumNumValues;
					}
					else
					{
						index = 0;
					}
					enumValue = selectedItem->enumValues[index];
				}
				else
				{
					enumValue = (enumValue + selectedItem->enumNumValues + enumChange) % 
						selectedItem->enumNumValues;
				}
			}

			lwlib_Edit_AccessItem(selectedItem, enum) = enumValue;
			break;
		case Edit_ItemType::Group:
			item_HandleGroupEditKeys(selectedItem);
			break;
		}

		if (priv.editMode == Edit_EditMode::Normal)
		{
			if (lwlib_Edit_KeyRepeat(NextItem))
			{
				cycleSelectedItem(1);
				priv.cursorCol = 0;
			}
			if (lwlib_Edit_KeyRepeat(PrevItem))
			{
				cycleSelectedItem(-1);
				priv.cursorCol = 0;
			}
		}
	}

	if (priv.editMode == Edit_EditMode::Normal)
	{
		if (lwlib_Edit_KeyRepeat(ExitGroup) && vec_size(priv.groupItemPath) > 0)
		{
			c = priv.nav->groupType == Edit_ItemGroupType::VecElem ? 2 : 1;
			for (i = 0; i < c; i++)
			{
				lwlib_StrCopy(priv.selectedItemName, vec_top(priv.groupItemPath).v);
				vec_pop_back(priv.groupItemPath);
			}
		}
	}

	if (forceExitStringMode)
	{
		priv.editMode = Edit_EditMode::Normal;
	}

	navigateTop();
}

void Edit::setActive(bool active)
{
	priv.active = active;
}

bool Edit::isActive()
{
	return priv.active;
}

void Edit::setFs(Fs *fs)
{
	priv.fs = fs;
}

void Edit::setFont(Font *font)
{
	priv.font = font;
}

void Edit::setKeys(lwlib_Keys_t keys)
{
	priv.keys = keys;
}

void Edit::addPersistGroup(Edit_Persist_t *persist)
{
	addItem_group("persist");
	addItem_string("fileName", 
		persist->fileName, sizeof(persist->fileName));
	addItem_enum("fileMode",
		(int *)&persist->fileMode, Fs_FileMode_Str,
		NULL, Fs_FileMode_StrCount);
	addItem_bool("commit", &persist->commit);
	navigateBack();
}

void Edit::setVecInit(const char *itemName, VecInit_t elemInit)
{
	Edit_Item_t *item;

	item = findItem(itemName);
	if (item != NULL)
	{
		item->groupVec->elemInit = elemInit;
	}
}

void Edit::setVecCleanup(const char *itemName, VecCleanup_t elemCleanup)
{
	Edit_Item_t *item;

	item = findItem(itemName);
	if (item != NULL)
	{
		item->groupVec->elemCleanup = elemCleanup;
	}
}

void Edit::setVecAddToEdit(const char *itemName, 
	VecAddToEdit_t elemAddToEdit)
{
	Edit_Item_t *item;

	item = findItem(itemName);
	if (item != NULL)
	{
		item->groupVec->elemAddToEdit = elemAddToEdit;
	}
}

void Edit::setItemEnabled(const char *itemName, bool enabled)
{
	setItemFlag(itemName, Edit_ItemFlag::Disabled, !enabled);
}
