/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#ifdef LEGACYENGINE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#endif
#include "lw.h"
#include "wl_def.h"
using namespace lwlib;

#ifdef LEGACYENGINE_OPENGL
namespace lwlib
{
	static int Gl_LoadTexture(uint8_t *texData, int texWidth, 
		int texHeight, int texBpp, bool mipmap, bool clamp)
	{
		int sw, sh;
		uint8_t *upData;
		GLuint glTex;
		GLint maxTexSize;

		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexSize);

		glGenTextures(1, &glTex);
		glBindTexture(GL_TEXTURE_2D, glTex);

		// set texture parameters
		if (!mipmap)
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
				GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
				GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
				GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
				GL_CLAMP_TO_EDGE);
		}
		else
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
				(mipmap ? GL_LINEAR_MIPMAP_NEAREST : GL_LINEAR));
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
				clamp ? GL_CLAMP_TO_EDGE : GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
				clamp ? GL_CLAMP_TO_EDGE : GL_REPEAT);
		}

		// scaled size, should be a power of 2, not greater then maxTexSize
		for (sw = 1; sw < texWidth; sw <<= 1);
		for (sh = 1; sh < texHeight; sh <<= 1);
		sw = MIN(sw, maxTexSize);
		sh = MIN(sh, maxTexSize);

		upData = lwlib_CallocMany(uint8_t, sw * sh * texBpp);
		if (sw != texWidth || sh != texHeight)
		{
			Img::resample(upData, sw, sh, texData, texWidth,
				texHeight, texBpp);
		}
		else
		{
			memcpy(upData, texData, sw * sh * texBpp);
		}

		// uploading base image
		glTexImage2D(GL_TEXTURE_2D, 0, texBpp, sw, sh, 0,
			texBpp == 4 ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE,
			upData);
		// uploading mipmaps if required
		if (mipmap)
		{
			int level = 0;

			while (Img::mipMap(upData, &sw, &sh, texBpp))
			{
				++level;
				glTexImage2D(GL_TEXTURE_2D, level, texBpp, sw, sh, 0,
					texBpp == 4 ? GL_RGBA : GL_RGB,
					GL_UNSIGNED_BYTE, upData);
			}
		}
		free(upData);

		return glTex;
	}
}
#endif

Gl::Gl(bool sdlEmu)
{
	lwlib_Zero(priv);

	if (sdlEmu)
	{
		priv.sdlGlEmu = new SDLGlEmu();
	}
}

Gl::~Gl()
{
#ifdef LEGACYENGINE_OPENGL
	int i;
	GLuint texId;
#endif

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		delete priv.sdlGlEmu;
	}

#ifdef LEGACYENGINE_OPENGL
	for (i = 0; (size_t)i < vec_size(priv.texIdSeq); i++)
	{
		texId = vec_at(priv.texIdSeq, i);
		glDeleteTextures(1, &texId);
	}
	vec_free(priv.texIdSeq, GLuint);
#endif
}

void Gl::setVideoMode(Gl_Screen_t *screen)
{
#ifdef LEGACYENGINE_OPENGL
	SDL_Surface *screenSurf;
#endif
	SDLGlEmu_Screen_t sdlGlEmuScreen;

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		lwlib_Zero(sdlGlEmuScreen);
		sdlGlEmuScreen.width = screen->width;
		sdlGlEmuScreen.height = screen->height;
		sdlGlEmuScreen.bits = screen->bits;
		sdlGlEmuScreen.fullscreen = screen->fullscreen;
		sdlGlEmuScreen.useDoubleBuffering = true;
#ifndef SDL_TRUE_COLOR
		sdlGlEmuScreen.palette = screen->palette;
#endif

		priv.sdlGlEmu->setVideoMode(&sdlGlEmuScreen);
		screen->screenSurf = sdlGlEmuScreen.screenSurf;
		screen->isDoubleBuffered = sdlGlEmuScreen.isDoubleBuffered;
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	screenSurf = SDL_SetVideoMode(screen->width, screen->height, screen->bits,
		(screen->fullscreen ? SDL_FULLSCREEN : 0) | SDL_OPENGL);
	if (!screenSurf)
	{
		printf("Unable to set video mode: %u x %u x %u bpp %s\n", 
			screen->width, screen->height, screen->bits, 
			"(double buffered)");
		exit(1);
		return;
	}

	screen->screenSurf = screenSurf;
	screen->isDoubleBuffered = true;

	priv.screenWidth = screen->width;
	priv.screenHeight = screen->height;
	priv.isDoubleBuffered = true;
#endif
}

void Gl::init()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->init();
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	/* viewport */
	glViewport(0, 0, screenWidth, screenHeight);
	glScissor(0, 0, screenWidth, screenHeight);

	/* set matrixes for 2D display */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, screenWidth, screenHeight, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* Misc Init */
	glClearColor(0.35, 0.35, 0.35, 0);
	glDisable(GL_LIGHTING);
	glDisable(GL_FOG);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glDisable(GL_DITHER);
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glPolygonMode(GL_FRONT, GL_FILL);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glCullFace(GL_BACK);

	/* Funcs Init */
	glDepthFunc(GL_LEQUAL);	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glAlphaFunc(GL_GREATER, 0.0);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* Hints */
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint(GL_FOG_HINT, GL_NICEST);

	/* clear buffers */
	glClear(GL_COLOR_BUFFER_BIT);
#endif
}

void Gl::clearScreen()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->clearScreen();
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	glClear(GL_COLOR_BUFFER_BIT);
#endif
}

void Gl::updateScreen()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->updateScreen();
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	SDL_GL_SwapBuffers();
#endif
}

int Gl::uploadTex(lwlib_Byte_t *data, int w, int h, int bpp,
	SDL_Color *palette)
{
	int texId;

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		return priv.sdlGlEmu->uploadTex(data, w, h, bpp,
			palette);
	}

#ifdef LEGACYENGINE_OPENGL
	texId = Gl_LoadTexture(data, w, h, bpp, false, true);
	vec_push(priv.texIdSeq, GLuint, texId);
#else
	texId = 0;
#endif

	return texId;
}

void Gl::deleteTex(int texId)
{
#ifdef LEGACYENGINE_OPENGL
	int i;
#endif

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->deleteTex(texId);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	assert(vec_size(priv.texIdSeq) > 0);

	for (i = 0; (size_t)i < vec_size(priv.texIdSeq); i++)
	{
		if (vec_at(priv.texIdSeq, i) == (size_t)texId)
		{
			break;
		}
	}

	assert((size_t)i != vec_size(priv.texIdSeq));
	vec_remove_unordered(priv.texIdSeq, i);

	glDeleteTextures(1, (GLuint *)&texId);
#endif
}

void Gl::lockScreen()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->lockScreen();
		return;
	}
}

void Gl::unlockScreen()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->unlockScreen();
		return;
	}
}

void Gl::drawScaledImage(int texId, int ix, int iy, int iw, int ih, 
	int w, int h, int cx, int cy, int sw, int sh, bool transparent)
{

#ifdef LEGACYENGINE_OPENGL
	struct
	{
		GLfloat xyz[3];
		GLfloat uv[2];
	} pts[4];
	GLfloat x1, x2, y1, y2;
	GLfloat u1, u2, v1, v2;
#endif

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->drawScaledImage(texId, ix, iy, 
			iw, ih, w, h, cx, cy, sw, sh, transparent);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	x1 = (GLfloat)cx;
	x2 = (GLfloat)(cx + sw);
	y1 = (GLfloat)cy;
	y2 = (GLfloat)(cy + sh);
	u1 = (GLfloat)ix / w;
	u2 = (GLfloat)(ix + iw) / w;
	v1 = (GLfloat)iy / h;
	v2 = (GLfloat)(iy + ih) / h;

	pts[0].xyz[0] = x1;
	pts[0].xyz[1] = y1;
	pts[0].xyz[2] = 0;
	pts[0].uv[0] = u1;
	pts[0].uv[1] = v1;
	pts[1].xyz[0] = x2;
	pts[1].xyz[1] = y1;
	pts[1].xyz[2] = 0;
	pts[1].uv[0] = u2;
	pts[1].uv[1] = v1;
	pts[2].xyz[0] = x2;
	pts[2].xyz[1] = y2;
	pts[2].xyz[2] = 0;
	pts[2].uv[0] = u2;
	pts[2].uv[1] = v2;
	pts[3].xyz[0] = x1;
	pts[3].xyz[1] = y2;
	pts[3].xyz[2] = 0;
	pts[3].uv[0] = u1;
	pts[3].uv[1] = v2;

	if (transparent)
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texId);

	glBegin(GL_QUADS);
	glTexCoord2fv(pts[0].uv);
	glVertex3fv(pts[0].xyz);
	glTexCoord2fv(pts[1].uv);
	glVertex3fv(pts[1].xyz);
	glTexCoord2fv(pts[2].uv);
	glVertex3fv(pts[2].xyz);
	glTexCoord2fv(pts[3].uv);
	glVertex3fv(pts[3].xyz);
	glEnd();

	glDisable(GL_TEXTURE_2D);

	if (transparent)
	{
		glDisable(GL_BLEND);
	}
#endif
}

void Gl::setCurColor(Point4b colorRgba)
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->setCurColor(colorRgba);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	glColor4ubv(colorRgba.v);
#endif
}

void Gl::resetCurColor()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->resetCurColor();
		return;
	}

	setCurColor(vec4b(0xFF, 0xFF, 0xFF, 0xFF));
}

void Gl::drawBox(int x, int y, int w, int h, Point4b colorRgba,
	bool lines)
{
#ifdef LEGACYENGINE_OPENGL
	const bool transparent = (lwlib_W(colorRgba) != 255);
#endif

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->drawBox(x, y, w, h, colorRgba, lines);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	if (lines)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}

	if (transparent)
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	setCurColor(colorRgba);

	glBegin(GL_QUADS);
	glVertex3f(x, y, 0);
	glVertex3f(x + w, y, 0);
	glVertex3f(x + w, y + h, 0);
	glVertex3f(x, y + h, 0);
	glEnd();

	resetCurColor();

	if (transparent)
	{
		glDisable(GL_BLEND);
	}

	if (lines)
	{
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
#endif
}

void Gl::drawBoxArrow(int x, int y, int w, int h, int r, 
	Point4b colorRgba, double dir)
{
#ifdef LEGACYENGINE_OPENGL
	const bool transparent = (lwlib_W(colorRgba) != 255);
#endif

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->drawBoxArrow(x, y, w, h, r, colorRgba, dir);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	if (transparent)
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	setCurColor(colorRgba);

	glPushMatrix();
	glTranslatef(x, y, 0);
	glRotatef(-dir * 90.0, 0, 0, 1);
	glTranslatef(r, 0, 0);
	glScalef(w, h, 1);

	glBegin(GL_TRIANGLES);
	glVertex3f(0, -1, 0);
	glVertex3f(1, 0, 0);
	glVertex3f(0, 1, 0);
	glEnd();

	glPopMatrix();

	resetCurColor();

	if (transparent)
	{
		glDisable(GL_BLEND);
	}
#endif
}

void Gl::pushMatrix()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->pushMatrix();
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	glPushMatrix();
#endif
}

void Gl::popMatrix()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->popMatrix();
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	glPopMatrix();
#endif
}

void Gl::translateMatrix(double tx, double ty, double tz)
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->translateMatrix(tx, ty, tz);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	glTranslated(tx, ty, tz);
#endif
}

void Gl::drawLine(int x1, int y1, int x2, int y2, 
	Point4b colorRgba)
{
#ifdef LEGACYENGINE_OPENGL
	const bool transparent = (lwlib_W(colorRgba) != 255);
#endif

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		priv.sdlGlEmu->drawLine(x1, y1, x2, y2, colorRgba);
		return;
	}

#ifdef LEGACYENGINE_OPENGL
	if (transparent)
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	setCurColor(colorRgba);

	glBegin(GL_LINES);
	glVertex3f(x1, y1, 0);
	glVertex3f(x2, y2, 0);
	glEnd();

	resetCurColor();

	if (transparent)
	{
		glDisable(GL_BLEND);
	}
#endif
}

int Gl::getWidth()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		return priv.sdlGlEmu->getWidth();
	}

	return priv.screenWidth;
}

int Gl::getHeight()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		return priv.sdlGlEmu->getHeight();
	}

	return priv.screenHeight;
}

Point2i Gl::getDims()
{

	if (lwlib_ObjValid(priv.sdlGlEmu))
	{
		return priv.sdlGlEmu->getDims();
	}

	return vec2i(priv.screenWidth, priv.screenHeight);
}

#if 0
void lwlib_GL_SaveBMP(SDL_Surface *surf, const char *fname)
{
	int i;
	int sw, sh;
	byte *grabPixels;
	int alloc;
	int pitch;

	sw = screenWidth;
	sh = screenHeight;
	pitch = screenPitch;

	grabAlloc = pitch * (sh + 1);
	grabPixels = (GLvoid *)malloc(grabAlloc);

	glReadPixels(0, 0, sw, sh, GL_RGBA, GL_UNSIGNED_BYTE, grabPixels);

	// flip image upside down
	for (i = 0; i < sh / 2; i++)
	{
		memcpy(&grabPixels[lwgl.grabAlloc - pitch],
			&grabPixels[i * pitch], pitch);
		memcpy(&grabPixels[i * pitch], &grabPixels[(sh - i - 1) * pitch],
			pitch);
		memcpy(&grabPixels[(sh - i - 1) * pitch], 
			&grabPixels[lwgl.grabAlloc - pitch], pitch);
	}

	surf = SDL_CreateRGBSurfaceFrom(lwgl.grabPixels, 
		sw, sh, 32, pitch, 
		lwgl.rmask, lwgl.gmask, lwgl.bmask, lwgl.amask);
	SaveBMPSDL(surf, fname);

	SDL_FreeSurface(surf);
}
#endif
