/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <stdio.h>
#include "lw_octree.h"
using namespace lwlib;

#define lwlib_Octree_MaxSizeLog2 16

#define lwlib_Octree_EachChildNode(x, node) (x) = 0; (size_t)(x) < lwlib_CountArray((node)->childNodes); (x)++
#define lwlib_Octree_EachChildNodeRev(x, node) (x) = lwlib_CountArray((node)->childNodes) - 1; (x) >= 0; (x)--

namespace lwlib
{
	static Octree_Node_t *createTopNode(int size)
	{
		Octree_Node_t *newNode;

		newNode = lwlib_CallocSingle(Octree_Node_t);
		newNode->size = size;
		newNode->cellBox = box3i(vec3i_zero(), vec3i(size, size, size));
		newNode->cellBoxCenter = vec3i(size / 2, size / 2, size / 2);

		return newNode;
	}

	static void createChildNode(Octree_Node_t *parentNode, int childNodeIndex)
	{
		int halfSize;
		Point3i childNodePoint;
		Point3i lower, upper;
		Octree_Node_t *newNode;

		assert(parentNode->size > 1 && lwlib_IntegerIsPow2(parentNode->size));

		halfSize = parentNode->size / 2;
		childNodePoint = vec_from_bits<int, 3>(childNodeIndex);
		lower = vec_add(lwlib_box_lower(parentNode->cellBox), 
			vec_scale(childNodePoint, halfSize));
		upper = vec_add(lower, 
			vec3i(halfSize, halfSize, halfSize));

		newNode = lwlib_CallocSingle(Octree_Node_t);
		newNode->size = halfSize;
		newNode->cellBox = box3i(lower, upper);
		newNode->cellBoxCenter = vec_divide(vec_add(lower, upper), 2);
		newNode->cellState = parentNode->cellState;
		newNode->parentNode = parentNode;
		newNode->childNodeIndex = childNodeIndex;

		parentNode->childNodes[childNodeIndex] = newNode;
	}

	static void initSideNodes(Octree_Node_t *node)
	{
		int i, k;
		#define CHILD(x) (x)
		#define SIDE(x) (-((x) + 1))
		static int s_tab[8][6] =
		{
			{ CHILD(1), SIDE(1), CHILD(2), SIDE(3), CHILD(4), SIDE(5) },
			{ SIDE(0), CHILD(0), CHILD(3), SIDE(3), CHILD(5), SIDE(5) },
			{ CHILD(3), SIDE(1), SIDE(2), CHILD(0), CHILD(6), SIDE(5) },
			{ SIDE(0), CHILD(2), SIDE(2), CHILD(1), CHILD(7), SIDE(5) },
			{ CHILD(5), SIDE(1), CHILD(6), SIDE(3), SIDE(4), CHILD(0) },
			{ SIDE(0), CHILD(4), CHILD(7), SIDE(3), SIDE(4), CHILD(1) },
			{ CHILD(7), SIDE(1), SIDE(2), CHILD(4), SIDE(4), CHILD(2) },
			{ SIDE(0), CHILD(6), SIDE(2), CHILD(5), SIDE(4), CHILD(3) },
		};

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, node))
			{
				lwlib_RepeatArray(node->childNodes[i]->sideNodes, j, sideNode,
					k = s_tab[i][j]; 
					sideNode = (k >= 0 ? node->childNodes[k] : node->sideNodes[-(k + 1)])
					);
				initSideNodes(node->childNodes[i]);
			}
		}
		#undef SIDE
		#undef CHILD
	}

	static Octree_Node_t *copyNode(Octree_Node_t *node)
	{
		int i;
		Octree_Node_t *newNode;

		newNode = lwlib_CallocSingle(Octree_Node_t);
		*newNode = *node;

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, node))
			{
				newNode->childNodes[i] = 
					copyNode(node->childNodes[i]);
			}
		}

		return newNode;
	}

	static void freeNode(Octree_Node_t *node)
	{
		int i;

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNodeRev(i, node))
			{
				freeNode(node->childNodes[i]);
			}
		}
		free(node);
	}

	static void freeChildNodes(Octree_Node_t *node)
	{
		int i;

		for (lwlib_Octree_EachChildNodeRev(i, node))
		{
			freeNode(node->childNodes[i]);
		}
		lwlib_Zero(node->childNodes);
	}

	static Octree_Node_t *iterFirstCell(Octree_Node_t *rootNode)
	{
		Octree_Node_t *curNode;

		curNode = rootNode;
		while (curNode && curNode->isParent)
		{
			curNode = curNode->childNodes[0];
		}
		return curNode;
	}

	static Octree_Node_t *iterNextCell(Octree_Node_t *rootNode,
		Octree_Node_t *curNode)
	{
		while (curNode != rootNode && 
			curNode->childNodeIndex + 1 == 
			lwlib_CountArray(curNode->childNodes))
		{
			curNode = curNode->parentNode;
		}
		if (curNode == rootNode)
		{
			return NULL;
		}

		curNode = 
			curNode->parentNode->childNodes[curNode->childNodeIndex + 1];
		while (curNode && curNode->isParent)
		{
			curNode = curNode->childNodes[0];
		}

		return curNode;
	}

	static bool findCellState(Octree_Node_t *node, bool cellState)
	{
		Octree_Node_t *curNode;

		curNode = iterFirstCell(node);
		while (curNode)
		{
			if (curNode->cellState == cellState)
			{
				return true;
			}
			curNode = iterNextCell(node, curNode);
		}
		return false;
	}

	static Octree_Node_t *childNodeForPos(Octree_Node_t *node, Point3i cellPos)
	{
		int childNodeIndex;
		childNodeIndex = vec_sign_bits(
			vec_sub(cellPos, node->cellBoxCenter), 0);
		lwlib_AssertArrayRange(childNodeIndex, node->childNodes);
		return node->childNodes[childNodeIndex];
	}

	static void setCellState(Octree_Node_t *node, Box3i cellBox, bool cellState)
	{
		int i;

		if (!box_touch(cellBox, node->cellBox))
		{
			return;
		}

		if (box_contains(cellBox, node->cellBox))
		{
			if (node->isParent)
			{
				freeChildNodes(node);
				node->isParent = false;
			}
			node->cellState = cellState;
		}
		else
		{
			if (!node->isParent && cellState != node->cellState)
			{
				node->isParent = true;
				for (lwlib_Octree_EachChildNode(i, node))
				{
					createChildNode(node, i);
				}

				initSideNodes(node);
			}

			if (node->isParent)
			{
				for (lwlib_Octree_EachChildNode(i, node))
				{
					setCellState(node->childNodes[i], 
						cellBox, cellState);
				}

				if (!findCellState(node, !cellState))
				{
					freeChildNodes(node);
					node->isParent = false;
					node->cellState = cellState;
				}
			}
		}
	}

	static void setCellStateCsg(Octree_Node_t *node, Csg *csg, bool cellState)
	{
		int i;
		Box3f box;

		box = box3i_float(node->cellBox);
		if (!csg->testBox(box, OverlapTest::Touch))
		{
			return;
		}

		if (csg->testBox(box, OverlapTest::Containment))
		{
			if (node->isParent)
			{
				freeChildNodes(node);
				node->isParent = false;
			}
			node->cellState = cellState;
		}
		else
		{
			if (!node->isParent && cellState != node->cellState && 
				node->size > 1)
			{
				node->isParent = true;
				for (lwlib_Octree_EachChildNode(i, node))
				{
					createChildNode(node, i);
				}

				initSideNodes(node);
			}

			if (node->isParent)
			{
				for (lwlib_Octree_EachChildNode(i, node))
				{
					setCellStateCsg(node->childNodes[i], 
						csg, cellState);
				}

				if (!findCellState(node, !cellState))
				{
					freeChildNodes(node);
					node->isParent = false;
					node->cellState = cellState;
				}
			}
		}
	}

	static bool queryCellState(Octree_Node_t *node, Point3i cellPos)
	{
		if (!box_test(node->cellBox, cellPos))
		{
			return false;
		}

		if (node->isParent)
		{
			return queryCellState(childNodeForPos(node, cellPos), cellPos);
		}

		return node->cellState;
	}

	static void negateNode(Octree_Node_t *node)
	{
		int i;

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, node))
			{
				negateNode(node->childNodes[i]);
			}
		}
		else
		{
			node->cellState = !node->cellState;
		}
	}

	static void applyNodeSetOp(Octree_Node_t *lNode,
		Octree_Node_t *rNode, Octree_SetOp::e setOp)
	{
		int i;

		if (lNode->isParent && rNode->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, lNode))
			{
				applyNodeSetOp(lNode->childNodes[i], 
					rNode->childNodes[i], setOp);
			}

			for (i = 0; i < 2; i++)
			{
				if (!findCellState(lNode, (bool)i))
				{
					freeChildNodes(lNode);
					lNode->isParent = false;
					lNode->cellState = (bool)(1 - i);
				}
			}
		}
		else if (lNode->isParent && !rNode->isParent)
		{
			if ((setOp == Octree_SetOp::Union && 
				rNode->cellState == true) ||
				(setOp == Octree_SetOp::Intersection &&
				rNode->cellState == false))
			{
				freeChildNodes(lNode);
				lNode->isParent = false;
				lNode->cellState = rNode->cellState;
			}
		}
		else if (!lNode->isParent && rNode->isParent)
		{
			if ((setOp == Octree_SetOp::Union && 
				lNode->cellState == false) ||
				(setOp == Octree_SetOp::Intersection &&
				lNode->cellState == true))
			{
				lNode->isParent = true;
				for (lwlib_Octree_EachChildNode(i, lNode))
				{
					lNode->childNodes[i] = 
						copyNode(rNode->childNodes[i]);
					initSideNodes(lNode);
				}
			}
		}
		else
		{
			assert(!lNode->isParent && !rNode->isParent);
			if (setOp == Octree_SetOp::Union)
			{
				lNode->cellState = 
					lNode->cellState || rNode->cellState;
			}
			else if (setOp == Octree_SetOp::Intersection)
			{
				lNode->cellState = 
					lNode->cellState && rNode->cellState;
			}
		}
	}

	static bool scaleNode(Octree_Node_t *node, int level)
	{
		int i;
		int shift;

		if (level == 0)
		{
			return false;
		}

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, node))
			{
				if (scaleNode(node->childNodes[i], level))
				{
					break;
				}
			}
			if (i != lwlib_CountArray(node->childNodes))
			{
				freeChildNodes(node);
				node->isParent = false;
				node->cellState = false;
			}
		}

		if (level > 0)
		{
			shift = level;
			node->size <<= shift;
			for (i = 0; i < 2; i++)
			{
				node->cellBox.p[i] = vec_op_scalar(node->cellBox.p[i], 
					VectorOp::ShiftLeft, shift);
			}
			node->cellBoxCenter = vec_op_scalar(node->cellBoxCenter, 
				VectorOp::ShiftLeft, shift);
		}
		else
		{
			shift = -level;
			node->size >>= shift;
			for (i = 0; i < 2; i++)
			{
				node->cellBox.p[i] = vec_op_scalar(node->cellBox.p[i], 
					VectorOp::ShiftRight, shift);
			}
			node->cellBoxCenter = vec_op_scalar(node->cellBoxCenter,
				VectorOp::ShiftRight, shift);

			if (node->size == 0)
			{
				return true;
			}
		}

		return false;
	}

	static void renderSurfaceVisitNode(Octree_Node_t *node, 
		Box3i cellBox, int facing, lwlib_Render_t render)
	{
		int i, j, k;
		Box3i clipBox;

		if (node == NULL)
		{
			render.drawBoxFacing(render.user, box3i_float(cellBox), facing);
			return;
		}

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, node))
			{
				clipBox = cellBox;
				for (j = 0; j < 3; j++)
				{
					k = 1 - lwlib_BitOn(i, j);
					lwlib_C(clipBox.p[k], j) = lwlib_MINMAX(
						lwlib_C(node->cellBoxCenter, j), 
						lwlib_C(clipBox.p[k], j), 1 - k);
					if (lwlib_C(clipBox.p[0], j) >= lwlib_C(clipBox.p[1], j))
					{
						break;
					}
				}
				if (j != 3)
				{
					continue;
				}
				renderSurfaceVisitNode(node->childNodes[i], 
					clipBox, facing, render);
			}
		}
		else
		{
			if (node->cellState != true)
			{
				render.drawBoxFacing(render.user, box3i_float(cellBox), facing);
			}
		}
	}

	static int countNodes(Octree_Node_t *node)
	{
		int i;
		int count = 1;

		if (node->isParent)
		{
			for (lwlib_Octree_EachChildNode(i, node))
			{
				count += countNodes(node->childNodes[i]);
			}
		}

		return count;
	}
}

void Octree::renderSolidBox(lwlib_Render_t render)
{
	Octree_Node_t *node;

	node = iterFirstCell(priv.topNode);
	while (node)
	{
		if (node->cellState == true)
		{
			render.drawBox(render.user, box3i_float(node->cellBox));
		}
		node = iterNextCell(priv.topNode, node);
	}
}

void Octree::renderSurface(lwlib_Render_t render)
{
	int facing;
	Octree_Node_t *node;

	node = iterFirstCell(priv.topNode);
	while (node)
	{
		if (node->cellState == true)
		{
			for (facing = 0; facing < 6; facing++)
			{
				if (priv.cullingEnabled && 
					!box3i_face_visible(node->cellBox, facing, 
						priv.cullingViewPos))
				{
					// no action
				}
				else
				{
					renderSurfaceVisitNode
						(
							node->sideNodes[facing],
							box_surface_facing(node->cellBox, facing),
							-(facing + 1), render
						);
				}
			}
		}
		node = iterNextCell(priv.topNode, node);
	}
}

void Octree::renderCells(lwlib_Render_t render)
{
	Point3i startPos, iterCur, iterEnd, cellPos;

	startPos = vec3i_zero();
	iterCur = vec3i_zero();
	iterEnd = vec3i_same(priv.topNode->size);

	if (priv.sliceEnabled)
	{
		lwlib_C(startPos, priv.sliceAxis) = priv.slicePos;
		lwlib_C(iterEnd, priv.sliceAxis) = 1;
	}

	do
	{
		cellPos = vec_add(startPos, iterCur);
		if (queryState(cellPos))
		{
			render.drawBox(render.user, box_cell(vec3i_float(cellPos)));
		}
		iterCur = vec_bounded_inc(iterCur, iterEnd);
	} while (!vec_equal(iterCur, vec3i_zero(), 0));
}

static void debugNodes(Octree_Node_t *node)
{
	int i;
	lwlib_String_t tmpstr[5];

	if (node->isParent)
	{
		printf("parent node size=%d, cellBox=%s, cellBoxCenter=%s\n",
			node->size, 
			box_to_string(tmpstr[0], "%d", node->cellBox), 
			vec_to_string(tmpstr[1], "%d", node->cellBoxCenter));
	}
	else
	{
		printf("child node size=%d, cellBox=%s, cellBoxCenter=%s, " \
			"cellState=%s, childNodeIndex=%d\n", 
			node->size, 
			box_to_string(tmpstr[0], "%d", node->cellBox), 
			vec_to_string(tmpstr[1], "%d", node->cellBoxCenter),
			lwlib_BoolString(node->cellState),
			node->childNodeIndex);
	}

	if (node->isParent)
	{
		for (lwlib_Octree_EachChildNode(i, node))
		{
			printf("entering child %d\n", i);
			debugNodes(node->childNodes[i]);
			printf("leaving child %d\n", i);
		}
	}
}

Octree::Octree(Octree_Node_t *topNode)
{
	lwlib_Zero(priv.topNode);
	lwlib_Zero(priv.sliceAxis);
	lwlib_Zero(priv.slicePos);
	lwlib_Zero(priv.sliceEnabled);
	lwlib_Zero(priv.cullingEnabled);
	lwlib_Zero(priv.cullingViewPos);

	priv.topNode = topNode;
}

Octree::Octree(int size)
{
	assert(size > 0 && lwlib_IntegerIsPow2(size));

	lwlib_Zero(priv.topNode);
	lwlib_Zero(priv.sliceAxis);
	lwlib_Zero(priv.slicePos);
	lwlib_Zero(priv.sliceEnabled);
	lwlib_Zero(priv.cullingEnabled);
	lwlib_Zero(priv.cullingViewPos);

	priv.topNode = createTopNode(size);
}

Octree::~Octree()
{
	freeNode(priv.topNode);
}

Octree *Octree::copy()
{
	Octree_Node_t *topNode;

	topNode = copyNode(priv.topNode);
	initSideNodes(topNode);

	return new Octree(topNode);
}

void Octree::expand(int size)
{
	Point3i cellPos, endCellPos;
	Octree *newOctree;

	assert(size > 0 && lwlib_IntegerIsPow2(size) && size > priv.topNode->size);

	newOctree = new Octree(size);
	cellPos = vec3i_zero();
	endCellPos = vec3i_same(priv.topNode->size);

	do
	{
		if (queryState(cellPos))
		{
			newOctree->setCellState(cellPos, true);
		}
		cellPos = vec_bounded_inc(cellPos, endCellPos);
	} while (!vec_equal(cellPos, vec3i_zero(), 0));

	lwlib_Swap(priv.topNode, newOctree->priv.topNode, Octree_Node_t *);
	delete newOctree;
}

void Octree::setState(Box3i cellBox, bool cellState)
{
	lwlib::setCellState(priv.topNode, cellBox, cellState);
}

void Octree::setCellState(Point3i cellPos, bool cellState)
{
	setState(box_cell(cellPos), cellState);
}

void Octree::setCsgState(Csg *csg, bool cellState)
{
	setCellStateCsg(priv.topNode, csg, cellState);
}

void Octree::applyBrush(Octree *brushOctree, bool cellState)
{
	Octree_Node_t *brushNode;

	assert(priv.topNode->size == brushOctree->priv.topNode->size);

	brushNode = iterFirstCell(brushOctree->priv.topNode);
	while (brushNode)
	{
		if (brushNode->cellState == true)
		{
			setState(brushNode->cellBox, cellState);
		}
		brushNode = iterNextCell(brushOctree->priv.topNode, brushNode);
	}
}

void Octree::applyNegate()
{
	negateNode(priv.topNode);
}

void Octree::applySetOp(Octree *rOctree, Octree_SetOp::e setOp)
{
	assert(priv.topNode->size == rOctree->priv.topNode->size);
	applyNodeSetOp(priv.topNode, rOctree->priv.topNode, setOp);
}

bool Octree::queryState(Point3i cellPos)
{
	return queryCellState(priv.topNode, cellPos);
}

void Octree::scale(int level)
{
	assert(priv.topNode->size >= 1 && 
		lwlib_IntegerIsPow2(priv.topNode->size));

	if 
		(
			lwlib_IntegerInRange(
				lwlib_Log2(priv.topNode->size) + level, 
				0, lwlib_Octree_MaxSizeLog2 + 1)
		)
	{
		scaleNode(priv.topNode, level);
	}
}

void Octree::render(Octree_RenderAlgo::e algo, lwlib_Render_t render)
{

	switch (algo)
	{
	case Octree_RenderAlgo::SolidBox:
		renderSolidBox(render);
		break;
	case Octree_RenderAlgo::Surface:
		renderSurface(render);
		break;
	case Octree_RenderAlgo::Cells:
		renderCells(render);
		break;
	}
}

void Octree::setSlice(bool sliceEnabled, int sliceAxis, int slicePos)
{
	priv.sliceEnabled = sliceEnabled;
	priv.sliceAxis = sliceAxis;
	priv.slicePos = slicePos;
}

int Octree::countNodes()
{
	return lwlib::countNodes(priv.topNode);
}

void Octree::debugPrint()
{
	debugNodes(priv.topNode);
}

int Octree::size()
{
	return priv.topNode->size;
}

void Octree::setCullingEnabled(bool cullingEnabled)
{
	priv.cullingEnabled = cullingEnabled;
}

void Octree::setCullingViewPos(Point3i cullingViewPos)
{
	priv.cullingViewPos = cullingViewPos;
}
