/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include "lw.h"

/*
=============================================================================

                                 RENDERING PLAN

=============================================================================
*/

// x-axis points straight out of the camera
// y-axis points left of the camera
// z-azis points directly up from the camera

// Roads span space infinitely along the y-axis. Quads (finite) will be used
// to model them. The length of quad along the y-axis increases as the quad is
// positioned further along the x-axis. The exact start and end position of
// quad along the y-axis is determined by the camera's position along y-axis
// and horizontal field-of-view. The limit of quad warping along y-axis of
// both polarities is taken into consideration when sizing quad. Once (u,v)
// are determined quad warping is applied. The quad is then projected onto
// screen space and rasterised.

// A. plot quads without warping (qx,qy,qz)
// B. derive texture coordinates (u,v) for all points (qx,qy,qz)
// C. apply warping to (qx,qy,qz) to obtain (px,py,pz) keeping same (u,v)
// D. apply rotation on (px,py,pz) to obtain (x,y,z)
// E. project (x,y,z) onto near plane (s,t)
// F. rasterise projected quads using (s,t) and (u,v)

// (qx,qy,qz) has no warping and no rotation
// (px,py,pz) has warping but no rotation
// (x,y,z) is (px,py,pz) after rotation

// warping functions f(x) and g(x) are available
//   px = qx
//   py = qy + f(qx)
//   pz = qz + g(qx)

// rotation transform
//   x = (pz - Cz) * sin(pitch) + (px - Cx) * cos(pitch)
//   y = py
//   z = (pz - Cz) * cos(pitch) - (px - Cx) * sin(pitch)

// (x,y,z) projection transform to (s,t) using parameters (near,vw,sw,vh,sh)
//   s = (0.5 - (y * near / x) / vw) * sw
//     = half(sw) - (y / x) * (near * sw / vw) ... (1)
//   t = (0.5 - (z * near / x) / vh) * sh
//     = half(sh) - (z / x) * (near * sh / vh) ... (2)

// let Pf be the far plane distance and Pn the near plane distance

// let D be a small distance, small enough to give the illusion of smooth
// motion

// f(x) and g(x) are only defined for x values from the sequence X
//   X = seq{ x : x = Pn + i * D and 0 <= i <= N }
// where N = (Pf - Pn) / D

// we have finite fifos F and G for f(x) and g(x), respectively
//   F = fifo{ f(Pn), f(Pn + D), ..., f(Pf) }
//   G = fifo{ g(Pn), g(Pn + D), ..., g(Pf) }

// the illusion of driving can be achieved with fifo pop/push on F and G

// the illusion also requires advancing the u coordinate with distance
// travelled

// camera follows warp contour according to warp values at the Pn plane
//   Cy = f(Pn)
//   Cz = g(Pn)

/*
=============================================================================

                                 RASTERISATION

=============================================================================
*/

#define SUBDIVSIZE 8
#define SUBDIVSHIFT 3

typedef struct lwlib_Raster_Edge_s
{
	const lwlib_Raster_Vertex_t *vertexs;
	int count;
	lwlib_FixedFrac_t xfrac;
	lwlib_FloatFrac_t uizfrac, vizfrac, izfrac;
} lwlib_Raster_Edge_t;

static inline void lwlib_Raster_InitEdge(lwlib_Raster_Edge_t *e,
	int cur, int next)
{
	e->count = e->vertexs[next].y - e->vertexs[cur].y;
	e->xfrac = lwlib_FixedFracFromInt(e->vertexs[cur].x,
		e->vertexs[next].x - e->vertexs[cur].x, e->count);
	e->uizfrac = lwlib_FloatFracInit(e->vertexs[cur].uiz,
		e->vertexs[next].uiz - e->vertexs[cur].uiz, e->count);
	e->vizfrac = lwlib_FloatFracInit(e->vertexs[cur].viz,
		e->vertexs[next].viz - e->vertexs[cur].viz, e->count);
	e->izfrac = lwlib_FloatFracInit(e->vertexs[cur].iz,
		e->vertexs[next].iz - e->vertexs[cur].iz, e->count);
}

#ifndef SDL_TRUE_COLOR
static void lwlib_Raster_DrawPolyFast2D(const lwlib_Raster_DrawPoly_t *
	drawPoly)
{
	int x, y;
	int x1, y1, x2, y2;
	fixedpt u, v;
	fixedpt u1, u2, v1, v2;
	lwlib_Color_Pix_t *vbuf, *vbufPtr;

	const lwlib_Raster_Vertex_t *vertexs = drawPoly->vertexs;
	const int vx = drawPoly->viewport.x;
	const int vy = drawPoly->viewport.y;
	const int vw = drawPoly->viewport.w;
	const int vh = drawPoly->viewport.h;
	const int pitch = drawPoly->screen.pitch;
	const int nu = drawPoly->texture.nu;
	const int ulim = nu - 1;
	const int nv = drawPoly->texture.nv;
	const int vlim = nv - 1;
	const int nvLog2 = drawPoly->texture.nvLog2;
	const lwlib_Color_Pix_t *tex = drawPoly->texture.tex;

	x1 = vertexs[0].x;
	y1 = vertexs[0].y;
	x2 = vertexs[1].x;
	y2 = vertexs[1].y;

	if (x1 >= vx + vw || x2 <= vx || x1 == x2 ||
		y1 >= vy + vh || y2 <= vy || y1 == y2)
	{
		return;
	}

	u1 = FIXEDPT_ONE * vertexs[0].uiz;
	v1 = FIXEDPT_ONE * vertexs[0].viz;
	u2 = FIXEDPT_ONE * vertexs[1].uiz;
	v2 = FIXEDPT_ONE * vertexs[1].viz;

	const fixedpt du = (u2 - u1) / (y2 - y1);
	const fixedpt dv = (v2 - v1) / (x2 - x1);

	if (x1 < vx)
	{
		v1 += dv * (vx - x1);
		x1 = vx;
	}

	if (x2 > vx + vw)
	{
		x2 = vx + vw;
	}

	if (y1 < vy)
	{
		u1 += du * (vy - y1);
		y1 = vy;
	}

	if (y2 > vy + vh)
	{
		y2 = vy + vh;
	}

	const int bufoffset = x1 + y1 * pitch;
	vbuf = drawPoly->screen.vbuf + bufoffset;

	const int nx = x2 - x1;
	const int ny = y2 - y1;

	u = u1;
	for (y = 0; y < ny; y++)
	{
		v = v1;
		vbufPtr = vbuf;
		for (x = 0; x < nx; x++)
		{
			const int uc = fixedpt_toint(u) & ulim;
			const int vc = fixedpt_toint(v) & vlim;

			const lwlib_Byte_t col = tex[(uc << nvLog2) + vc];
			if (col != 0xff)
			{
				*vbufPtr = col;
			}

			v += dv;
			vbufPtr++;
		}

		u += du;
		vbuf += pitch;
	}
}
#endif

void lwlib_Raster_DrawPoly(const lwlib_Raster_DrawPoly_t *drawPoly)
{
#define PREV(i) (((i) + n - 1) % n)
#define NEXT(i) (((i) + 1) % n)
	int x, y;
	fixedpt u, v, w;
	fixedpt u1, u2, v1, v2, w1, w2;
	int lx, rx;
	fixedpt du, dv, dw;
	int nx;
	int cur, next, prev;
	int left, right;
	int minX, maxX, minY, maxY;
	lwlib_Color_Pix_t *vbuf, *vbufPtr, *vbufPtrHigh;
	fixedpt *zbuf, *zbufPtr;
	lwlib_FloatFrac_t uizfrac, vizfrac, izfrac;
	lwlib_Raster_Edge_t ledge, redge;
	float z;
	lwlib_Color_Pix_t texCol;
#ifdef SDL_TRUE_COLOR
	int i;
	int blend[4];
#endif
	float uizdxn, vizdxn, izdxn;

#ifndef SDL_TRUE_COLOR
	const bool fast2D = drawPoly->fast2D;
	if (fast2D)
	{
		lwlib_Raster_DrawPolyFast2D(drawPoly);
		return;
	}
#endif

	const lwlib_Raster_Vertex_t *vertexs = drawPoly->vertexs;
	const int n = drawPoly->numVertexs;
	const int vx = drawPoly->viewport.x;
	const int vy = drawPoly->viewport.y;
	const int vw = drawPoly->viewport.w;
	const int vh = drawPoly->viewport.h;
	const int pitch = drawPoly->screen.pitch;
	const int nu = drawPoly->texture.nu;
	const int ulim = nu - 1;
	const int nv = drawPoly->texture.nv;
	const int vlim = nv - 1;
	const int nvLog2 = drawPoly->texture.nvLog2;
	const lwlib_Color_Pix_t *tex = drawPoly->texture.tex;
	//const lwlib_Byte_t *shadetable = drawPoly->shadetable;
	const bool affine = drawPoly->affine;

#ifdef SDL_TRUE_COLOR
	const uint32_t blendPix = drawPoly->blendPix;
	for (i = 0; i < 4; i++)
	{
		blend[i] = ((lwlib_Byte_t *)&blendPix)[i] << 8;
	}
#endif

	ledge.vertexs = redge.vertexs = vertexs;

	minX = vx + vw;
	maxX = vx;

	minY = vy + vh;
	maxY = vy;

	for (cur = 0; cur < n; cur++)
	{
		minX = lwlib_MIN(vertexs[cur].x, minX);
		maxX = lwlib_MAX(vertexs[cur].x, maxX);
		minY = lwlib_MIN(vertexs[cur].y, minY);
		maxY = lwlib_MAX(vertexs[cur].y, maxY);
	}

	if (minX >= vx + vw || maxX <= vx || minX == maxX ||
		minY >= vy + vh || maxY <= vy || minY == maxY)
	{
		return;
	}

	for (cur = 0; cur < n; cur++)
	{
		next = NEXT(cur);
		prev = PREV(cur);

		if (vertexs[cur].y < vertexs[prev].y && 
			vertexs[cur].y <= vertexs[next].y)
		{
			break;
		}
	}
	if (cur == n)
	{
		return;
	}
	left = cur;
	prev = PREV(left);
	lwlib_Raster_InitEdge(&ledge, left, prev);

	right = left;
	while (vertexs[right].y == vertexs[left].y)
	{
		right = NEXT(right);
	}
	right = PREV(right);
	next = NEXT(right);
	lwlib_Raster_InitEdge(&redge, right, next);

	const int bufoffset = minY * pitch;
	vbuf = drawPoly->screen.vbuf + bufoffset;
	zbuf = drawPoly->screen.zbuf + bufoffset;

	for (y = minY; y < maxY; y++, vbuf += pitch, zbuf += pitch)
	{
		lx = lwlib_FixedFracToInt(&ledge.xfrac);
		rx = lwlib_FixedFracToInt(&redge.xfrac);
		if (y >= vy && y < vy + vh && rx > vx && lx < vx + vw && lx < rx)
		{
			nx = rx - lx;
			uizfrac = lwlib_FloatFracInit(ledge.uizfrac.val, 
				redge.uizfrac.val - ledge.uizfrac.val, nx);
			vizfrac = lwlib_FloatFracInit(ledge.vizfrac.val, 
				redge.vizfrac.val - ledge.vizfrac.val, nx);
			izfrac = lwlib_FloatFracInit(ledge.izfrac.val, 
				redge.izfrac.val - ledge.izfrac.val, nx);

			x = lx;
			vbufPtr = vbuf + x;
			zbufPtr = zbuf + x;

			if (x < vx)
			{
				const int nc = vx - x;
				lwlib_FloatFracIncMany(&uizfrac, nc);
				lwlib_FloatFracIncMany(&vizfrac, nc);
				lwlib_FloatFracIncMany(&izfrac, nc);
				nx -= nc;
				vbufPtr += nc;
				zbufPtr += nc;
				x = vx;
			}

			if (x + nx > vx + vw)
			{
				nx = vx + vw - x;
			}

			if (affine)
			{
				z = FIXEDPT_ONE;
			}
			else
			{
				z = FIXEDPT_ONE / izfrac.val;

				uizdxn = uizfrac.inc * SUBDIVSIZE;
				vizdxn = vizfrac.inc * SUBDIVSIZE;
				izdxn = izfrac.inc * SUBDIVSIZE;
			}

			u2 = uizfrac.val * z;
			v2 = vizfrac.val * z;
			w2 = FIXEDPT_ONE * izfrac.val;

			while (nx > 0)
			{
				if (!affine && nx >= SUBDIVSIZE)
				{
					uizfrac.val += uizdxn;
					vizfrac.val += vizdxn;
					izfrac.val += izdxn;

					u1 = u2;
					v1 = v2;
					w1 = w2;

					z = FIXEDPT_ONE / izfrac.val;
					u2 = uizfrac.val * z;
					v2 = vizfrac.val * z;
					w2 = FIXEDPT_ONE * izfrac.val;

					u = u1;
					v = v1;
					w = w1;

					du = (u2 - u1) >> SUBDIVSHIFT;
					dv = (v2 - v1) >> SUBDIVSHIFT;
					dw = (w2 - w1) >> SUBDIVSHIFT;
					vbufPtrHigh = vbufPtr + SUBDIVSIZE;
					nx -= SUBDIVSIZE;
				}
				else
				{
					uizfrac.val += uizfrac.inc * nx;
					vizfrac.val += vizfrac.inc * nx;
					izfrac.val += izfrac.inc * nx;

					u1 = u2;
					v1 = v2;
					w1 = w2;

					if (!affine)
					{
						z = FIXEDPT_ONE / izfrac.val;
					}
					u2 = uizfrac.val * z;
					v2 = vizfrac.val * z;
					w2 = FIXEDPT_ONE * izfrac.val;

					u = u1;
					v = v1;
					w = w1;

					du = (u2 - u1) / nx;
					dv = (v2 - v1) / nx;
					dw = (w2 - w1) / nx;
					vbufPtrHigh = vbufPtr + nx;
					nx = 0;
				}

				while (vbufPtr < vbufPtrHigh)
				{
					if (w >= *zbufPtr)
					{
						const int uc = fixedpt_toint(u) & ulim;
						const int vc = fixedpt_toint(v) & vlim;

						texCol = tex[(uc << nvLog2) + vc];
#ifdef SDL_TRUE_COLOR
						lwlib_Color_BlendColor((lwlib_Byte_t *)&texCol,
							blend);
						lwlib_Color_BlendAlpha((lwlib_Byte_t *)&texCol, 
							(lwlib_Byte_t *)vbufPtr);
						//*zbufPtr = w;
#else
						if (texCol != 0xff)
						{
							*vbufPtr = texCol;
							*zbufPtr = w;
						}
#endif

						u += du;
						v += dv;
						w += dw;
						vbufPtr++;
						zbufPtr++;
					}
					else
					{
						const lwlib_Color_Pix_t *vbufPtrHidden = vbufPtr;
						do
						{
							w += dw;
							vbufPtr++;
							zbufPtr++;
						} while (vbufPtr < vbufPtrHigh && w < *zbufPtr);

						const int hcount = vbufPtr - vbufPtrHidden;
						u += du * hcount;
						v += dv * hcount;
					}
				}
			}
		}

		lwlib_FixedFracInc(&ledge.xfrac);
		lwlib_FloatFracInc(&ledge.uizfrac);
		lwlib_FloatFracInc(&ledge.vizfrac);
		lwlib_FloatFracInc(&ledge.izfrac);
		ledge.count--;
		while (ledge.count == 0)
		{
			left = PREV(left);
			prev = PREV(left);
			if (vertexs[prev].y == vertexs[left].y)
			{
				continue;
			}
			lwlib_Raster_InitEdge(&ledge, left, prev);
		}

		lwlib_FixedFracInc(&redge.xfrac);
		lwlib_FloatFracInc(&redge.uizfrac);
		lwlib_FloatFracInc(&redge.vizfrac);
		lwlib_FloatFracInc(&redge.izfrac);
		redge.count--;
		while (redge.count == 0)
		{
			right = NEXT(right);
			next = NEXT(right);
			if (vertexs[next].y == vertexs[right].y)
			{
				continue;
			}
			lwlib_Raster_InitEdge(&redge, right, next);
		}
	}
#undef NEXT
#undef PREV
}

