/*************************************************************************
** Wolf3D Legacy
** Copyright (C) 2012 by LinuxWolf - Team RayCAST
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**************************************************************************
** LinuxWolf Library for Wolf3D Legacy
*************************************************************************/

#include "lw_tileviewer.h"
#include "lw_img.h"
#include "lw_intmap.h"
#include "wl_def.h"
using namespace lwlib;

#define LWLIB_TILEVIEWER_BORDER 0.015
#define LWLIB_TILEVIEWER_MAXACROSS 20
#define LWLIB_TILEVIEWER_MAXDOWN 15
#define LWLIB_TILEVIEWER_RED 4

#define lwlib_TileViewer_KeyAction(name) lwlib_TileViewer_KeyAction_ ## name
#define lwlib_TileViewer_KeyAction_MoveUp "k"
#define lwlib_TileViewer_KeyAction_MoveDown "j"
#define lwlib_TileViewer_KeyAction_MoveLeft "h"
#define lwlib_TileViewer_KeyAction_MoveRight "l"
#define lwlib_TileViewer_KeyAction_NextPage "n"
#define lwlib_TileViewer_KeyAction_PrevPage "b"
#define lwlib_TileViewer_KeyAction_JumpToSprites "s"
#define lwlib_TileViewer_KeyAction_JumpToWalls "w"
#define lwlib_TileViewer_KeyPressed(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_TileViewer_KeyAction(name), lwlib_Keys_Event(Pressed))
#define lwlib_TileViewer_KeyReleased(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_TileViewer_KeyAction(name), lwlib_Keys_Event(Released))
#define lwlib_TileViewer_KeyRepeat(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_TileViewer_KeyAction(name), lwlib_Keys_Event(Repeat))
#define lwlib_TileViewer_KeyFastRepeat(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_TileViewer_KeyAction(name), lwlib_Keys_Event(FastRepeat))
#define lwlib_TileViewer_KeyDown(name) lwlib_Keys_CheckEvent(priv.keys, \
	lwlib_TileViewer_KeyAction(name), lwlib_Keys_Event(Down))

typedef lwlib_Byte_t byte;

static void lwlib_TileViewer_ShapeToImage(int shapenum, byte *image)
{
	int x, y, i;
	word *cmdptr;
	t_compshape *shape;
	byte *sprite;
	short *linecmds;

	memset(image, 255, TEXTURESIZE * TEXTURESIZE);

	shape = (t_compshape *)PM_GetSprite(shapenum);
	sprite = (byte *)shape;

	cmdptr = shape->dataofs;
	for (x = shape->leftpix; x <= shape->rightpix; x++)
	{
		linecmds = (short *)(sprite + *cmdptr++);
		for (; *linecmds; linecmds += 3)
		{
			i = linecmds[2] / 2 + linecmds[1];
			for (y = linecmds[2] / 2; y < linecmds[0] / 2; y++, i++)
			{
				image[x * 64 + y] = sprite[i];
			}
		}
	}
}

ED_Sprite_t *TileViewer::selectedSprite()
{
	int spriteNum;
	ED_Sprite_t *sprite;
	ED_SpriteSeq_t *sprites;

	if (priv.tileSelector >= PMSpriteStart && priv.tileSelector < PMSoundStart)
	{
		sprites = ED_Sprites(ed);

		spriteNum = priv.tileSelector - PMSpriteStart;
		if ((size_t)spriteNum < vec_size(*sprites))
		{
			sprite = &vec_at(*sprites, spriteNum);
			return sprite;
		}
	}

	return NULL;
}

void TileViewer::drawTileInfo()
{
	ED_Sprite_t *sprite;
	lwlib_String_t tmpStr;

	sprite = selectedSprite();
	if (sprite != NULL)
	{
		lwlib_Snprintf(tmpStr, "%d: \"%s\"", priv.tileSelector - PMSpriteStart, sprite->name);
	}
	else
	{
		lwlib_Snprintf(tmpStr, "%d", priv.tileSelector);
	}

	priv.font->printPosFit(vec2i(0, 0), tmpStr);
}

static void lwlib_TileViewer_StatInfoMaker_AddToEdit(
	TileViewer_StatInfoMaker_t *statInfoMaker, Edit *edit)
{
	edit->addItem_bool("doInsert", 
		&statInfoMaker->doInsert);
	edit->addItem_string("name", 
		statInfoMaker->name, sizeof(statInfoMaker->name));
	edit->addItem_enum("type", 
		(int *)&statInfoMaker->type, ED_Stat_Str, NULL,
		ED_Stat_StrCount);
	edit->addItem_string("bonus", 
		statInfoMaker->bonus, sizeof(statInfoMaker->bonus));
	edit->addItem_enum("specialFlags", 
		(int *)&statInfoMaker->specialFlags,
		ED_ObjFlag_Str, ED_ObjFlag_Val,
		ED_ObjFlag_StrCount);
	edit->addItem_int("animframes", 
		&statInfoMaker->animframes);
}

static void lwlib_TileViewer_StatInfoReport_AddToEdit(
	TileViewer_StatInfoReport_t *statInfoReport, Edit *edit)
{
	edit->addItem_bool("generateReport", 
		&statInfoReport->generateReport);
}

TileViewer::TileViewer()
{
	lwlib_Zero(priv);

	priv.tilesAcross = 4;
	priv.tilesDown = 1;
}

TileViewer::~TileViewer()
{
	int i;
	for (i = 0; (size_t)i < vec_size(priv.tileImgs); i++)
	{
		delete vec_at(priv.tileImgs, i);
	}
	vec_free(priv.tileImgs, Img_t);
	lwlib_IntMapClear(&priv.tileImgMap);
}

void TileViewer::addToEdit(Edit *edit)
{
	edit->addItem_int("tileSelector", &priv.tileSelector);
	edit->addItem_group("extraOptions");
	edit->addItem_int("tilesAcross", &priv.tilesAcross);
	edit->addItem_int("tilesDown", &priv.tilesDown);
	edit->addItem_bool("toSpriteStart", &priv.toSpriteStart);
	edit->addItem_bool("toWallStart", &priv.toWallStart);
	edit->navigateBack();
	lwlib_Edit_AddGroup(edit, "statInfoMaker", 
		lwlib_TileViewer_StatInfoMaker_AddToEdit(&priv.statInfoMaker, edit));
	lwlib_Edit_AddGroup(edit, "statInfoReport", 
		lwlib_TileViewer_StatInfoReport_AddToEdit(&priv.statInfoReport, edit));
}

void TileViewer::updateEdit(Edit *edit)
{
	ED_Sprite_t *sprite;
	ED_StatInfo_t statInfo;

	sprite = selectedSprite();
	if (sprite != NULL)
	{
		edit->addItem_string("spriteName", 
			sprite->name, sizeof(sprite->name));

		if (priv.statInfoMaker.doInsert)
		{
			priv.statInfoMaker.doInsert = false;

			lwlib_Zero(statInfo);
			lwlib_StrCopy(statInfo.name, priv.statInfoMaker.name);
			lwlib_StrCopy(statInfo.picnum, sprite->name);
			statInfo.type = priv.statInfoMaker.type;
			lwlib_StrCopy(statInfo.bonus, priv.statInfoMaker.bonus);
			statInfo.specialFlags = priv.statInfoMaker.specialFlags;
			statInfo.animframes = priv.statInfoMaker.animframes;
			ED_AddStatInfo(ed, statInfo);

			lwlib_StrCopy(ED_Inject(ed)->actorName, 
				priv.statInfoMaker.name);
		}

		if (priv.statInfoReport.generateReport)
		{
			priv.statInfoReport.generateReport = false;
			ED_StatInfoReport(ed);
		}
	}
}

void TileViewer::updateFrame()
{
	int tileSelector;
	int nw, nh;
	int tilesPerPage;
	int page, cell;

	if (priv.toSpriteStart)
	{
		priv.toSpriteStart = false;
		priv.tileSelector = PMSpriteStart;
	}

	if (priv.toWallStart)
	{
		priv.toWallStart = false;
		priv.tileSelector = 0;
	}

	if (priv.edit->isActive())
	{
		return;
	}

	tileSelector = priv.tileSelector;
	nw = lwlib_CLIP(priv.tilesAcross, 1, LWLIB_TILEVIEWER_MAXACROSS - 1);
	nh = lwlib_CLIP(priv.tilesDown, 1, LWLIB_TILEVIEWER_MAXDOWN - 1);
	tilesPerPage = nw * nh;
	page = tileSelector / tilesPerPage;
	cell = tileSelector % tilesPerPage;

	if (lwlib_TileViewer_KeyRepeat(MoveRight))
	{
		if ((cell % nw) == nw - 1)
		{
			tileSelector = (page + 1) * tilesPerPage + ((cell / nw) * nw);
		}
		else
		{
			tileSelector++;
		}
	}

	if (lwlib_TileViewer_KeyRepeat(MoveLeft))
	{
		if ((cell % nw) == 0)
		{
			tileSelector = (page - 1) * tilesPerPage + ((cell / nw) * nw + nw - 1);
		}
		else
		{
			tileSelector--;
		}
	}

	if (lwlib_TileViewer_KeyRepeat(MoveDown))
	{
		tileSelector += nw;
	}

	if (lwlib_TileViewer_KeyRepeat(MoveUp))
	{
		tileSelector -= nw;
	}

	if (lwlib_TileViewer_KeyRepeat(NextPage))
	{
		tileSelector += tilesPerPage;
	}

	if (lwlib_TileViewer_KeyRepeat(PrevPage))
	{
		tileSelector -= tilesPerPage;
	}

	if (lwlib_TileViewer_KeyPressed(JumpToSprites))
	{
		priv.toSpriteStart = true;
	}

	if (lwlib_TileViewer_KeyPressed(JumpToWalls))
	{
		priv.toWallStart = true;
	}

	priv.tileSelector = tileSelector;
}

void TileViewer::draw(Gl *gl)
{
	int x, y;
	int sw, sh, cx, cy;
	int nw, nh;
	int bw, bh;
	int px, py;
	int tileStart, tileRem;

	nw = lwlib_CLIP(priv.tilesAcross, 1, LWLIB_TILEVIEWER_MAXACROSS - 1);
	nh = lwlib_CLIP(priv.tilesDown, 1, LWLIB_TILEVIEWER_MAXDOWN - 1);

	sw = priv.gl->getWidth();
	sh = priv.gl->getHeight();

	bw = (int)(LWLIB_TILEVIEWER_BORDER * (double)sw);
	bh = (int)(LWLIB_TILEVIEWER_BORDER * (double)sh);

	cx = (sw / 2);
	cy = (sh / 2);

	sw = TEXTURESIZE;
	cx = cx - (sw + 2 * bw) * nw / 2;

	sh = TEXTURESIZE;
	cy = cy - (sh + 2 * bh) * nh / 2;

	tileStart = (priv.tileSelector / (nw * nh)) * (nw * nh);
	tileRem = priv.tileSelector % (nw * nh);

	for (x = 0; x < nw; x++)
	{
		for (y = 0; y < nh; y++)
		{
			px = cx + (sw + 2 * bw) * x + bw;
			py = cy + (sh + 2 * bh) * y + bh;

			if (y * nw + x == tileRem)
			{
				priv.gl->drawBox(px - bw, 
					py - bh, sw + 2 * bw, 
					sh + 2 * bh, GAMEPAL_RGBA(LWLIB_TILEVIEWER_RED, 0), 
					false);
			}

			drawTile(tileStart + y * nw + x, px, py, sw, sh);
		}
	}

	if (!priv.edit->isActive())
	{
		drawTileInfo();
	}
}

void TileViewer::setGl(Gl *gl)
{
	priv.gl = gl;
}

void TileViewer::setKeys(lwlib_Keys_t keys)
{
	priv.keys = keys;
}

void TileViewer::setFont(Font *font)
{
	priv.font = font;
}

void TileViewer::setPalette(SDL_Color *palette)
{
	priv.palette = palette;
}

void TileViewer::setEdit(Edit *edit)
{
	priv.edit = edit;
}

Img_t TileViewer::getTileImg(int tile)
{
    byte *source;
	Img_t img;

	lwlib_Zero(img);

	if (!lwlib_IntMapContains(&priv.tileImgMap, tile))
	{
		lwlib_Zero(img);
		if (tile < PMSpriteStart)
		{
			source = (byte *)PM_GetPage(tile);
			img = new Img();
			img->fromSource(source, TEXTURESIZE, TEXTURESIZE, 1, 
				TEXTURESIZE, false, gamepal);
		}
		else if (tile < PMSoundStart)
		{
			source = lwlib_CallocMany(byte, TEXTURESIZE * TEXTURESIZE);
			lwlib_TileViewer_ShapeToImage(tile - PMSpriteStart, source);
			img = new Img();
			img->fromSource(source, TEXTURESIZE, TEXTURESIZE, 1, 
				TEXTURESIZE, true, gamepal);
		}

		if (img == NULL)
		{
			return img;
		}

		lwlib_IntMapInsert(&priv.tileImgMap, tile, 
			(lwlib_Word_t)vec_size(priv.tileImgs));

		img->transpose();
#ifdef SDL_TRUE_COLOR
		img->expandPalette(true);
#endif

		vec_push(priv.tileImgs, Img_t, img);
	}
	else
	{
		img = vec_at(priv.tileImgs, 
			(int)lwlib_IntMapElem(&priv.tileImgMap, tile));
	}

	return img;
}

void TileViewer::drawTile(int tile, int cx, int cy, int sw, int sh)
{
	Gl *gl;
	Img_t img;

	gl = priv.gl;

	img = getTileImg(tile);
	if (img != NULL)
	{
		img->drawScaled(gl, cx, cy, sw, sh, false);
	}
}
